/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "../common/common.h"
#include "avi.h"

#define INDEXSTEP	1024

typedef struct aviindex
{
	int32_t Id;
	int8_t Flags[4];
	int32_t Pos;
	int32_t Length;
} aviindex;

typedef struct avistream
{
	format_stream Stream;
	
	int* Index;			// mediatime
	int RefStart;
	int MediaTime;
	int NewMediaTime;	// after seek (if seek successed)
	fraction MediaRate; // byterate or packetrate
	int MaxPacket;

} avistream;

typedef struct avi
{
	format_base Format;

	int FramePeriod;
	int ByteRate;
	bool_t OutOfSync;
	bool_t ErrorIndex;

	int DataPos;
	int DataEnd;

	// stream header and format loading
	int HeadNo;
	bool_t HeadType;

	int VideoHandler;
	int VideoFourCC;
	int VideoBitCount;
	int VideoFrames;
	fraction VideoFPS;
	int VideoWidth;
	int VideoHeight;
	int VideoStart;

	int AudioLength;
	fraction AudioRate;
	int AudioFormat;
	int AudioChannels;
	int AudioSampleRate;
	int AudioByteRate;
	int AudioBits;
	int AudioStart;
	int AudioSampleSize;

	int IndexCount;
	int IndexPos;
	int IndexReaded;
	int IndexBufferNo;
	aviindex IndexBuffer[INDEXSTEP];

	char NameChar[256];

} avi;

#define ALIGN2(x) (((x)+1)&~1)
#define ISDIGIT(x) ((x)>='0' && (x)<='9')
#define STREAMNO(x) ((((x) & 255)-'0')*10 + ((((x) >> 8) & 255)-'0'))
#define ISDATA(x) (ISDIGIT((x) & 255) && ISDIGIT(((x) >> 8) & 255))
#define ISIDX(x) (ISDIGIT(((x) >> 16) & 255) && ISDIGIT(((x) >> 24) & 255) && ((x) & 255)=='i' && (((x) >> 8) & 255)=='x')

static int Init(avi* p)
{
	p->IndexPos = -1;
	p->IndexCount = 0;
	p->IndexReaded = 0;
	p->IndexBufferNo = -1;
	p->DataPos = -1;
	p->DataEnd = -1;
	p->OutOfSync = 0;
	p->ErrorIndex = 0;

	memset(p->NameChar,0,sizeof(p->NameChar));
	memset(p->NameChar+'a',1,'z'-'a');
	memset(p->NameChar+'A',1,'Z'-'A');
	memset(p->NameChar+'0',1,'9'-'0');
	p->NameChar['_']=1;
	p->NameChar[' ']=1;

	return ERR_NONE;
}

static int Probe(avi* p, const void* Ptr, int Length, const tchar_t* URL )
{
	const uint8_t* Data = (const uint8_t*)Ptr;

    if (Data[0]=='R' &&	Data[1]=='I' &&	Data[2]=='F' &&	Data[3]=='F' &&
		Data[8]=='A' &&	Data[9]=='V' &&	Data[10]=='I'&&	Data[11]==' ')
		return ERR_NONE;

	if (CheckExts(URL,p->Format.Exts))
		return ERR_NONE;

	return ERR_INVALID_DATA;
}

INLINE int ChunkTime( avistream* Stream, int DataLength )
{
	if (Stream->Stream.Fragmented)
		return DataLength; //byterate

	// packetrate
	if (DataLength > Stream->MaxPacket) 
		return (DataLength + Stream->MaxPacket - 1) / Stream->MaxPacket;
	return 1;
}

static int ReadPacket(avi* p, format_packet* Packet)
{
	avistream* Stream;
	int No,Ch1,Ch2,StartPos;
	int SubType,Type,Length,DataLength;

	// find a valid chunk header first

	//DEBUG_MSG1("AVI Packet %08x",p->Format.FilePos);

	StartPos = p->Format.FilePos;

	do
	{
		No = 0;
		Type = 0;

		while (No<32)
		{
			Ch1 = Format_Read8(&p->Format);
			Ch2 = Format_Read8(&p->Format);

			if (Ch1>=0 && Ch2>=0 && p->NameChar[Ch1] && p->NameChar[Ch2])
			{
				Type |= (Ch1+(Ch2<<8)) << No;
				No += 16;
			}
			else
			{
				if (p->Format.NeedMoreInput)
					return ERR_NONE;

				p->OutOfSync = 1;
				if ((p->Format.FilePos-StartPos)>=CHUNKSIZE || p->Format.NeedMoreInput)
					return ERR_DATA_NOT_FOUND; // this will force the process loop to stop, so input thread can be wake up
				No = 0;
				Type = 0;
			}
		}

		if (p->DataEnd >= 0 && p->Format.FilePos >= p->DataEnd)
			return ERR_END_OF_FILE;

		DataLength = Format_ReadLE32(&p->Format);

		if (p->Format.NeedMoreInput)
			return ERR_NONE;

		if (DataLength<0 || (p->Format.FileSize>=0 && p->Format.FilePos+DataLength>p->Format.FileSize))
		{
			p->OutOfSync = 1;
			continue;
		}

		Length = ALIGN2(DataLength);

		if (ISDATA(Type))
		{
			No = STREAMNO(Type);
			if (No >= 0 && No < p->Format.StreamCount)
			{
				//DEBUG_MSG3("AVI Packet stream:%d pos:%08x length:%d",No,p->Format.FilePos,DataLength);

				if (p->OutOfSync)
				{
					if (p->Format.ProcessTime >= 0)
					{
						int i;
						tick_t RefTime;
						avistream* Stream;
	
						// this won't solve the out of sync problem
						// but at least the streams won't be dropped

						for (i=0;i<p->Format.StreamCount;++i)
						{
							Stream = (avistream*) p->Format.Streams[i];
							RefTime = Stream->RefStart+Scale(Stream->MediaTime,Stream->MediaRate.Num,Stream->MediaRate.Den);
							if (RefTime < p->Format.ProcessTime - TICKSPERSEC)
								Stream->MediaTime = Scale(p->Format.ProcessTime-Stream->RefStart,Stream->MediaRate.Den,Stream->MediaRate.Num);
						}
					}
					p->OutOfSync = 0;
				}

				Stream = (avistream*) p->Format.Streams[No];
				Packet->Stream = &Stream->Stream;
				Packet->Data = Format_ReadAsRef(&p->Format,DataLength);
				Format_Skip(&p->Format,Length - DataLength);
				Packet->RefTime = Stream->RefStart+Scale(Stream->MediaTime,Stream->MediaRate.Num,Stream->MediaRate.Den);

				//DEBUG_MSG5(T("AVI stream:%d time:%d mediatime:%d pos:%08x length:%08x"),No,Packet->RefTime,Stream->MediaTime,p->Format.FilePos,DataLength);

				if (!p->Format.NeedMoreInput)
					Stream->MediaTime += ChunkTime(Stream,DataLength);

				return ERR_NONE;
			}
		}
		else
		if (ISIDX(Type))
		{
			// nice to have: read OpenDML index  
		}
		else
		switch (Type)
		{
		case FOURCCLE('R','I','F','F'):
			if (p->Format.FileSize < 0 || p->Format.FileSize > Length + 8)
				p->Format.FileSize = Length + 8;
			p->HeadNo = -1;
			Format_ReadLE32(&p->Format); // 'AVI '
			return ERR_NONE;

		case FOURCCLE('a','v','i','h'):
			p->FramePeriod = Format_ReadLE32(&p->Format);
			p->ByteRate = Format_ReadLE32(&p->Format);
			Length -= 8;
			break;

		case FOURCCLE('s','t','r','h'):

			p->HeadType = Format_ReadLE32(&p->Format);
			Length -= 4;

			if (p->HeadType == FOURCCLE('v','i','d','s'))
			{
				p->VideoHandler = Format_ReadLE32(&p->Format);

				Format_ReadLE32(&p->Format); // flags
				Format_ReadLE16(&p->Format); // priority
				Format_ReadLE16(&p->Format); // language
				Format_ReadLE32(&p->Format); 

				p->VideoFPS.Den = Format_ReadLE32(&p->Format); 
				p->VideoFPS.Num = Format_ReadLE32(&p->Format); 

				if (!p->VideoFPS.Den || !p->VideoFPS.Num)
				{
					p->VideoFPS.Den = 1;
					p->VideoFPS.Num = 25;
				}

				p->VideoStart = Format_ReadLE32(&p->Format);
				p->VideoFrames = Format_ReadLE32(&p->Format);

				Length -= 32;
			}
			else
			if (p->HeadType == FOURCCLE('a','u','d','s'))
			{
				Format_ReadLE32(&p->Format); // format
				Format_ReadLE32(&p->Format); // flags
				Format_ReadLE16(&p->Format); // priority
				Format_ReadLE16(&p->Format); // language
				Format_ReadLE32(&p->Format); 
				p->AudioRate.Den = Format_ReadLE32(&p->Format); 
				p->AudioRate.Num = Format_ReadLE32(&p->Format); 
				p->AudioStart = Format_ReadLE32(&p->Format);
				p->AudioLength = Format_ReadLE32(&p->Format);

				Format_ReadLE32(&p->Format); // initialframes
				Format_ReadLE32(&p->Format); // suggested buffersize

				p->AudioSampleSize = Format_ReadLE32(&p->Format);

				Length -= 44;
			}
			break;

		case FOURCCLE('s','t','r','f'):
			if (p->HeadType == FOURCCLE('v','i','d','s'))
			{
				Format_ReadLE32(&p->Format); //size
				p->VideoWidth = Format_ReadLE32(&p->Format); 
				p->VideoHeight = Format_ReadLE32(&p->Format); 
				Format_ReadLE16(&p->Format); //planes
				p->VideoBitCount = Format_ReadLE16(&p->Format);
				p->VideoFourCC = Format_ReadLE32(&p->Format);

				Length -= 20;
				Format_Skip(&p->Format,Length);

				if (!p->Format.NeedMoreInput && p->HeadNo >= p->Format.StreamCount)
				{
					avistream* Stream = (avistream*) Format_AddStream(&p->Format,sizeof(avistream),sizeof(video));
					video* Format = (video*) Stream->Stream.Format;
					tick_t Duration;

					Stream->Index = NULL;
					Stream->Stream.FormatType = PACKET_VIDEO;
					Stream->Stream.Fragmented = 0;
					Stream->Stream.PacketBurst = VIDEO_BURST;

					Simplify(&p->VideoFPS,MAX_INT/TICKSPERSEC,MAX_INT);

					Stream->MediaRate.Den = p->VideoFPS.Num;
					Stream->MediaRate.Num = TICKSPERSEC * p->VideoFPS.Den;
					Stream->MaxPacket = MAX_INT;
					Stream->RefStart = Scale(p->VideoStart,Stream->MediaRate.Num,Stream->MediaRate.Den);

					Format->FrameRate = p->VideoFPS;

					Duration = Scale(p->VideoFrames,Stream->MediaRate.Num,Stream->MediaRate.Den);
					if (p->Format.Duration < Duration)
						p->Format.Duration = Duration;

					Format->Width = p->VideoWidth;
					Format->Height = p->VideoHeight;
					Format->Aspect = ASPECT_ONE;
					Format->Pixel.BitCount = p->VideoBitCount;

					if (!p->VideoFourCC) // DIB?
					{
						switch (p->VideoBitCount)
						{
						case 16:
							DefaultRGB(&Format->Pixel,p->VideoBitCount,5,5,5,0,0,0);
							break;
						case 24:
						case 32:
							DefaultRGB(&Format->Pixel,p->VideoBitCount,8,8,8,0,0,0);
							break;
						}
						Format->Direction = DIR_MIRRORUPDOWN;
					}
					else
					{
						Format->Pixel.Flags = PF_FOURCC;
						Format->Pixel.FourCC = UpperFourCC(p->VideoFourCC);
					}
					DefaultPitch(Format);

					if (Format->Height<0)
					{
						Format->Height = -Format->Height;
						Format->Direction ^= DIR_MIRRORUPDOWN;
					}
					return ERR_STREAM_ADDED;
				}
				return ERR_NONE;

			}
			else
			if (p->HeadType == FOURCCLE('a','u','d','s'))
			{
				p->AudioFormat = Format_ReadLE16(&p->Format);
				p->AudioChannels = Format_ReadLE16(&p->Format); 
				p->AudioSampleRate = Format_ReadLE32(&p->Format); 
				p->AudioByteRate = Format_ReadLE32(&p->Format); 
				Format_ReadLE16(&p->Format); // blockalign
				Length -= 14;
				if (p->AudioFormat == AUDIOFMT_PCM)
				{
					p->AudioBits = Format_ReadLE16(&p->Format);
					Length -= 2;
				}
		
				Format_Skip(&p->Format,Length);

				if (!p->Format.NeedMoreInput && p->HeadNo >= p->Format.StreamCount)
				{
					avistream* Stream = (avistream*) Format_AddStream(&p->Format,sizeof(avistream),sizeof(audio));
					audio* Format = (audio*) Stream->Stream.Format;
					tick_t Duration;

					Stream->Index = NULL;
					Stream->Stream.FormatType = PACKET_AUDIO;

					if (!p->AudioSampleSize)
					{
						// VBR (packetrate)
						Simplify(&p->AudioRate,MAX_INT/TICKSPERSEC,MAX_INT);
						Stream->Stream.Fragmented = 0;
						Stream->MediaRate.Den = p->AudioRate.Num;
						Stream->MediaRate.Num = TICKSPERSEC * p->AudioRate.Den;
						Stream->MaxPacket = p->AudioRate.Den;
					}
					else
					{
						// CBR (byterate)
						Stream->Stream.Fragmented = 1;

						if (p->AudioFormat == 0x55 /*MP3*/ && p->AudioRate.Num == 4977)
						{
							Stream->MediaRate.Den = (p->AudioRate.Num*16-7)*p->AudioSampleSize;
							Stream->MediaRate.Num = TICKSPERSEC*16*p->AudioRate.Den;
						}
						else
						{
							Stream->MediaRate.Den = p->AudioRate.Num*p->AudioSampleSize;
							Stream->MediaRate.Num = TICKSPERSEC*p->AudioRate.Den;
						}
					}

					Simplify(&p->AudioRate,MAX_INT,MAX_INT/TICKSPERSEC);
					Stream->RefStart = Scale(p->AudioStart,TICKSPERSEC*p->AudioRate.Den,p->AudioRate.Num);
					Duration = Scale(p->AudioLength,TICKSPERSEC*p->AudioRate.Den,p->AudioRate.Num);
					if (Duration>0 && p->Format.Duration < Duration)
						p->Format.Duration = Duration;

					Format->Format = p->AudioFormat;
					Format->Channels = p->AudioChannels;
					Format->SampleRate = p->AudioSampleRate;
					Format->ByteRate = p->AudioByteRate;
					Format->Bits = p->AudioBits;
					Format->FracBits = p->AudioBits - 1;
					if (Format->Bits <= 8)
						Format->Flags = PCM_UNSIGNED;

					return ERR_STREAM_ADDED;
				}
				return ERR_NONE;
			}
			break;

		case FOURCCLE('L','I','S','T'):
			SubType = Format_ReadLE32(&p->Format);
			Length -= 4;

			if (!p->Format.NeedMoreInput)
				switch (SubType)
				{
				case FOURCCLE('s','t','r','l'):
					p->HeadNo++;
					return ERR_NONE;

				case FOURCCLE('h','d','r','l'):
					return ERR_NONE;

				case FOURCCLE('m','o','v','i'):
					p->DataPos = p->Format.FilePos;
					p->DataEnd = Length + p->Format.FilePos;
					p->Format.PreLoad = PRELOAD; // no more preload needed

					for (No=0;No<p->Format.StreamCount;++No)
						((avistream*)p->Format.Streams[No])->MediaTime = 0;

					return ERR_NONE;
				}
			break;

		case FOURCCLE('s','t','r','n'):
		case FOURCCLE('i','n','d','x'):
		case FOURCCLE('J','U','N','K'):
			break;

		default: 
			if (p->OutOfSync)   // unknown
				Length = -1;
		}
	}
	while (Length<0);

	if (!p->Format.NeedMoreInput)
		Format_Skip(&p->Format,Length);

	return ERR_NONE;
}

static int ErrorIndex(avi* p)
{
	if (!p->ErrorIndex)
	{
		p->ErrorIndex = 1;
		NodeError((node*)p,AVI_ID,AVI_NOINDEX);
	}
	return ERR_NOT_SUPPORTED;
}

static int LoadIndexBuffer(avi* p,int No)
{
	if (p->IndexBufferNo != No)
	{
		int Bytes = (p->IndexCount - No*INDEXSTEP)*sizeof(aviindex);
		if (Bytes > INDEXSTEP*sizeof(aviindex))
			Bytes = INDEXSTEP*sizeof(aviindex);

		if (p->Format.Input->Seek(p->Format.Input,p->IndexPos+No*INDEXSTEP*sizeof(aviindex),SEEK_SET)<0)
			return ErrorIndex(p);

		if (p->Format.Input->Read(p->Format.Input,p->IndexBuffer,Bytes) != Bytes)
			return ErrorIndex(p);

		p->IndexBufferNo = No;
	}
	return ERR_NONE;
}

static int Seek(avi* p, tick_t Time, int FilePos )
{
	avistream* Stream;
	avistream* s;
	int i,j,No;
	int Result;

	if (p->DataPos < 0)
	{
		// no avi header yet, only seek to the beginning supported

		if (Time>0 || FilePos>0)
			return ERR_NOT_SUPPORTED;

		return Format_InputSeek(&p->Format,0);
	}

	// seek to the beginning? (no index needed for that)

	Stream = (avistream*) Format_DefSyncStream(&p->Format);

	if (Time <= 0 && FilePos <= 0)
	{
		if (Format_InputSeek(&p->Format,p->DataPos) != ERR_NONE)
			return ERR_NOT_SUPPORTED;

		for (j=0;j<p->Format.StreamCount;++j)
		{
			s = (avistream*) p->Format.Streams[j];
			s->MediaTime = 0;
		}

		p->OutOfSync = 0;
		return ERR_NONE;
	}

	// do we have the index block?

	if (p->IndexPos < 0)
	{
		int Pos = p->DataEnd;
		// try to find index block

		if (p->Format.Input->Seek(p->Format.Input,Pos,SEEK_SET) < 0)
			return ErrorIndex(p);

		for (;;)
		{
			int32_t Head[2];
			if (p->Format.Input->Read(p->Format.Input,Head,sizeof(Head)) != sizeof(Head))
				return ErrorIndex(p);

			Pos += sizeof(Head);

			if (Head[0] == FOURCC('i','d','x','1'))
			{
				p->IndexCount = Head[1] / sizeof(aviindex);

				i = ((p->IndexCount + INDEXSTEP - 1) / INDEXSTEP) + 1;

				for (No=0;No<p->Format.StreamCount;++No)
				{
					s = (avistream*) p->Format.Streams[No];
					s->Index = (int*) ReAlloc(s->Index,sizeof(int)*i);
					if (!s->Index)
						return ERR_OUT_OF_MEMORY;

					s->Index[0] = 0;
				}

				p->IndexPos = Pos;
				p->IndexReaded = 0;
				break;
			}

			Pos += ALIGN2(Head[1]);
			if (p->Format.Input->Seek(p->Format.Input,Pos,SEEK_SET) != Pos)
				return ErrorIndex(p);
		}
	}

	if (Stream)
	{
		int Next;
		int MediaTime;
		int IndexLeft = p->IndexCount;
		int IndexCount;
		int Pos;
		int OldFilePos = p->Format.Input->Seek(p->Format.Input,0,SEEK_CUR);
		bool_t NeedKey = Stream->Stream.FormatType == PACKET_VIDEO;

		MediaTime = Scale(Time,Stream->MediaRate.Den,Stream->MediaRate.Num);

		// nice to have: support seek by file position

		// find the index page

		for (No=0;IndexLeft>0;++No,IndexLeft-=INDEXSTEP)
		{
			IndexCount = IndexLeft;
			if (IndexCount > INDEXSTEP)
				IndexCount = INDEXSTEP;

			if (p->IndexReaded <= No)
			{
				// load page and calc sum mediatime

				Result = LoadIndexBuffer(p,No);
				if (Result != ERR_NONE)
				{
					p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
					return Result;
				}

				for (j=0;j<p->Format.StreamCount;++j)
				{
					s = (avistream*) p->Format.Streams[j];
					s->Index[No+1] = s->Index[No];
				}

				for (i=0;i<IndexCount;++i)
				{
					j = STREAMNO(p->IndexBuffer[i].Id);
					if (j>=0 && j<p->Format.StreamCount)
					{
						s = (avistream*) p->Format.Streams[j];
						s->Index[No+1] += ChunkTime(s,INT32LE(p->IndexBuffer[i].Length));
					}
				}

				p->IndexReaded = No+1;
			}

			//DEBUG_MSG4("AVI IndexBlock no:%d start:%d end:%d (target:%d)",No,Stream->Index[No],Stream->Index[No+1],MediaTime);

			if (Stream->Index[No] <= MediaTime && Stream->Index[No+1] > MediaTime)
			{
				Result = LoadIndexBuffer(p,No);
				if (Result != ERR_NONE)
					return Result;

				for (j=0;j<p->Format.StreamCount;++j)
				{
					s = (avistream*) p->Format.Streams[j];
					s->NewMediaTime = s->Index[No];
				}

				// find sub page position

				for (i=0;i<IndexCount;++i)
				{
					j = STREAMNO(p->IndexBuffer[i].Id);
					if (j>=0 && j<p->Format.StreamCount)
					{
						s = (avistream*) p->Format.Streams[j];

						Next = s->NewMediaTime + ChunkTime(s,p->IndexBuffer[i].Length);
	
						//DEBUG_MSG3("AVI Index stream:%d pos:%d flag:%d",j,INT32LE(p->IndexBuffer[i].Pos),p->IndexBuffer[i].Flags[0]);

						if (s==Stream && MediaTime < Next)
						{
							Pos = INT32LE(p->IndexBuffer[i].Pos)-4;

							// only accept keyframes with valid position
							if (Pos >= 0 && Pos < p->DataEnd - p->DataPos &&
								(!NeedKey || (p->IndexBuffer[i].Flags[0] & 0x10)))
								break;
							MediaTime = Next;
						}

						s->NewMediaTime = Next;
					}
				}
				
				// found? 
				if (i<IndexCount)
					break;

				// continue on next page
				MediaTime = Stream->Index[No+1];
			}
		}

		if (IndexLeft <= 0)
		{
			// go to end
			for (j=0;j<p->Format.StreamCount;++j)
			{
				s = (avistream*) p->Format.Streams[j];
				s->NewMediaTime = s->Index[p->IndexReaded];
			}

			Pos = p->DataEnd - p->DataPos;
		}

		if (Format_InputSeek(&p->Format,p->DataPos+Pos) == ERR_NONE)
		{
			// set new positions
			for (j=0;j<p->Format.StreamCount;++j)
			{
				s = (avistream*) p->Format.Streams[j];
				s->MediaTime = s->NewMediaTime;
			}

			p->OutOfSync = 0;
			return ERR_NONE;
		}

		// restore old position
		p->Format.Input->Seek(p->Format.Input,OldFilePos,SEEK_SET);
	}
	return ERR_NOT_SUPPORTED;
}

void FreeStream(avi* p, avistream* Stream)
{
	Free(Stream->Index);
}

static avi AVI;

void AVI_Init()
{
	AVI.Format = FormatBase;
	AVI.Format.VMT.Probe = Probe;
	AVI.Format.Init = Init;
	AVI.Format.Seek = Seek;
	AVI.Format.ReadPacket = ReadPacket;
	AVI.Format.FreeStream = FreeStream;
	AVI.Format.Id = AVI_ID;
	AVI.Format.Exts = T("avi:V;divx:V");

	NodeRegister((node*)&AVI.Format.VMT,PRI_DEFAULT);
}

void AVI_Done()
{
	NodeUnRegister((node*)&AVI.Format.VMT);
}