/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"
#include "../dyncode/dyncode.h"
#include "blit_soft.h"

#define SAT(Value) Value = Value < 0 ? 0: (Value > 255 ? 255: Value);

static rgb Gray1[2] = { 
	{0,0,0},{255,255,255}
};
static rgb Gray2[4] = { 
	{0,0,0},{85,85,85},{170,170,170},{255,255,255} 
};
static rgb Gray4[16] = { 
	{0,0,0},{17,17,17},{34,34,34},{51,51,51},
	{68,68,68},{85,85,85},{102,102,102},{119,119,119},
	{136,136,136},{153,153,153},{170,170,170},{187,187,187},
	{204,204,204},{221,221,221},{238,238,238},{255,255,255}
};

typedef struct blitpack
{
	blitfx FX;
	video Dst;
	video Src;
	rect DstRect;
	rect SrcRect;
	blit_soft Code[2];
	struct blitpack* Next;
	int RScaleX;
	int RScaleY;
	int SrcAlignPos;
	int DstAlignPos;
	int DstAlignSize;

} blitpack;

static blitpack* BlitPack = NULL;

void Blit_Init()
{
}

void Blit_Done()
{
	blitpack* q;
	blitpack* p;
	
	for (p=BlitPack;p;p=q)
	{
		q = p->Next;
		CodeDone(&p->Code[0].Code);
		CodeDone(&p->Code[1].Code);
		Free(p);
	}

	BlitPack = NULL;
}

static const rgb* DefaultPal( const pixel* Format )
{
	if (Format->Flags & PF_PALETTE)
	{
		if (!Format->Palette)
			switch (Format->BitCount)
			{
			case 1: return Gray1;
			case 2: return Gray2;
			case 4: return Gray4;
			}
		return Format->Palette;
	}
	return NULL;
}


int CalcScale( int v, int Min, int Max )
{
	if (v > Max) v = Max;
	if (v < Min) v = Min;
	return v;
}

int CalcRScale( int v, int Gray )
{
	if (v<=0) return 16;

	v = (16*1024 << 16) / v;

	if (Gray) // only 100% and 200% scale
		return v > 12288 ? 16:8;

	//align to 100%
	if (v > 16834-1024 && v < 16384+1024) 
		v = 16384;
	//align to 200%
	if (v > 8192-1024 && v < 8192+1024) 
		v = 8192;
	//align to 50%
	if (v > 32768-1024 && v < 32768+1024) 
		v = 32768;

#if defined(MIPS) || defined(SH3)
	return ((v+1024) >> 11) << 1;
#else
	return (v+512) >> 10;
#endif
}

void* BlitCreate( const video* Dst, 
			      const video* Src, const blitfx* FX )
{
	bool_t Gray;

	blitpack* p = BlitPack;
	if (p)
		BlitPack = p->Next;
	else
	{
		p = (blitpack*) CAlloc(sizeof(blitpack),1);
		if (!p)
			return NULL;

		CodeInit(&p->Code[0].Code);
		CodeInit(&p->Code[1].Code);
	}

	if (!BlitCompile(&p->Code[0],Dst,Src,FX,0) ||
		!BlitCompile(&p->Code[1],Dst,Src,FX,1))
	{
		p->Next = BlitPack;
		BlitPack = p;
		return NULL;
	}

	p->FX = *FX;
	p->Dst = *Dst;
	p->Src = *Src;

	Gray = (Dst->Pixel.Flags & PF_PALETTE) && 
		   (Dst->Pixel.BitCount == 4 || Dst->Pixel.BitCount == 2);

	p->RScaleX = CalcRScale(FX->ScaleX,Gray);
	p->RScaleY = CalcRScale(FX->ScaleY,Gray);

	// default alignment
	p->SrcAlignPos = 2;
	p->DstAlignPos = 2;
	p->DstAlignSize = 2;

	if (Gray && Dst->Pixel.BitCount == 2) // arm_gray, mips_gray, sh3_gray 
		p->DstAlignPos = p->DstAlignSize = 4;
#ifdef ARM
	else
	if (PlanarYUV(&Dst->Pixel)) // arm_yuv
		p->SrcAlignPos = p->DstAlignPos = p->DstAlignSize = 8;
	else 
	if (p->RScaleX == 32 && p->RScaleY == 32) //arm_half
		p->DstAlignPos = p->DstAlignSize = 4;
	else 
	if ((p->RScaleX == 8 || p->RScaleY == 8) && //arm_fix with double 
		(p->RScaleX == 8 || p->RScaleX == 16) && 
		(p->RScaleY == 8 || p->RScaleY == 16) && !(p->FX.Flags & (BLITFX_DITHER|BLITFX_ARITHSTRETCH)))
		p->DstAlignSize = 4;
#endif

	return p;
}

void BlitAlign( void* Handle, rect* DstRect, rect* SrcRect )
{
	int i;
	int ShrinkX,ShrinkY;
	int SrcRight;
	int SrcBottom;
	blitpack* p = Handle;
	int RScaleX,RScaleY;

	if (!p) return;

	RScaleX = p->RScaleX;
	RScaleY = p->RScaleY;

	SrcRight = SrcRect->x + SrcRect->Width;
	SrcBottom = SrcRect->y + SrcRect->Height;
	
	// convert source to destination space
	if (p->FX.Direction & DIR_SWAPXY)
	{
		SwapInt(&RScaleX,&RScaleY);
		SwapRect(SrcRect);
	}

	SrcRect->Width = SrcRect->Width * 16 / RScaleX;
	SrcRect->Height = SrcRect->Height * 16 / RScaleY;

	ShrinkX = DstRect->Width - SrcRect->Width;
	if (ShrinkX>0) //shrink destination?
	{
		DstRect->x += (DstRect->Width - SrcRect->Width) >> 1;
		DstRect->Width = SrcRect->Width;
	}
	else //adjust source position
		SrcRect->x += (SrcRect->Width - DstRect->Width) * RScaleX >> 5;
	
	ShrinkY = DstRect->Height - SrcRect->Height;
	if (ShrinkY>0) //shrink Dst?
	{
		DstRect->y += (DstRect->Height - SrcRect->Height) >> 1;
		DstRect->Height = SrcRect->Height;
	}
	else //adjust source position
		SrcRect->y += (SrcRect->Height - DstRect->Height) * RScaleY >> 5;

	i = DstRect->x & (p->DstAlignPos-1);
	if (i && (ShrinkX >> 1) < i)
	{
		DstRect->Width -= p->DstAlignPos - i;
		DstRect->x += p->DstAlignPos - i;
	}
	else
		DstRect->x -= i;

	i = DstRect->y & (p->DstAlignPos-1);
	if (i && (ShrinkY >> 1) < i)
	{
		DstRect->Height -= p->DstAlignPos - i;
		DstRect->y += p->DstAlignPos - i;
	}
	else
		DstRect->y -= i;

	DstRect->Width &= ~(p->DstAlignSize-1); 
	DstRect->Height &= ~(p->DstAlignSize-1);

	SrcRect->x &= ~(p->SrcAlignPos-1);
	SrcRect->y &= ~(p->SrcAlignPos-1);

	// convert source back to it's space

	SrcRect->Width = (DstRect->Width * RScaleX / 16) & ~1;
	SrcRect->Height = (DstRect->Height * RScaleY / 16) & ~1;

	if (p->FX.Direction & DIR_SWAPXY)
		SwapRect(SrcRect);

	if (SrcRect->x + SrcRect->Width > SrcRight)
		SrcRect->Width = SrcRight - SrcRect->x;

	if (SrcRect->y + SrcRect->Height > SrcBottom)
		SrcRect->Height = SrcBottom - SrcRect->y;

	p->DstRect = *DstRect;
	p->SrcRect = *SrcRect;
}

void BlitRelease(void* Handle)
{
	blitpack* p = (blitpack*)Handle;
	if (p)
	{
		p->Next = BlitPack;
		BlitPack = p;
	}
}

int AnyAlign( rect* DstRect, rect* SrcRect, const blitfx* FX, 
			  int DstAlignSize, int DstAlignPos,
			  int MinScale, int MaxScale )
{
	int i,ShrinkX,ShrinkY;
	int ScaleX,ScaleY;
	int SrcRight;
	int SrcBottom;

	if (!DstRect || !SrcRect || !FX)
		return ERR_INVALID_PARAM;

	ScaleX = CalcScale(FX->ScaleX,MinScale,MaxScale);
	ScaleY = CalcScale(FX->ScaleY,MinScale,MaxScale);
	SrcRight = SrcRect->x + SrcRect->Width;
	SrcBottom = SrcRect->y + SrcRect->Height;

	// convert source to destination space
	if (FX->Direction & DIR_SWAPXY)
	{
		SwapInt(&ScaleX,&ScaleY);
		SwapRect(SrcRect);
	}

	SrcRect->Width = (SrcRect->Width * ScaleX + 32768 ) >> 16;
	SrcRect->Height = (SrcRect->Height * ScaleY + 32768 ) >> 16;

	ShrinkX = DstRect->Width - SrcRect->Width;
	if (ShrinkX>0) //shrink destination?
	{
		DstRect->x += (DstRect->Width - SrcRect->Width) >> 1;
		DstRect->Width = SrcRect->Width;
	}
	else //adjust source position
		SrcRect->x += ((SrcRect->Width - DstRect->Width) << 15) / ScaleX;
	
	ShrinkY = DstRect->Height - SrcRect->Height;
	if (ShrinkY>0) //shrink Dst?
	{
		DstRect->y += (DstRect->Height - SrcRect->Height) >> 1;
		DstRect->Height = SrcRect->Height;
	}
	else //adjust source position
		SrcRect->y += ((SrcRect->Height - DstRect->Height) << 15) / ScaleY;

	// final alignment

	i = DstRect->x & (DstAlignPos-1);
	if (i && (ShrinkX >> 1) < i)
	{
		DstRect->Width -= DstAlignPos - i;
		DstRect->x += DstAlignPos - i;
	}
	else
		DstRect->x -= i;

	i = DstRect->y & (DstAlignPos-1);
	if (i && (ShrinkY >> 1) < i)
	{
		DstRect->Height -= DstAlignPos - i;
		DstRect->y += DstAlignPos - i;
	}
	else
		DstRect->y -= i;

	DstRect->Width &= ~(DstAlignSize-1); 
	DstRect->Height &= ~(DstAlignSize-1);

	SrcRect->x &= ~1;
	SrcRect->y &= ~1;

	SrcRect->Width = ((DstRect->Width << 16) / ScaleX) & ~1;
	SrcRect->Height = ((DstRect->Height << 16) / ScaleY) & ~1;

	if (FX->Direction & DIR_SWAPXY)
		SwapRect(SrcRect);

	if (SrcRect->x + SrcRect->Width > SrcRight)
		SrcRect->Width = SrcRight - SrcRect->x;

	if (SrcRect->y + SrcRect->Height > SrcBottom)
		SrcRect->Height = SrcBottom - SrcRect->y;

	return ERR_NONE;
}

INLINE void SurfacePtr( uint8_t** Ptr, const planes Planes, const video* Format, int BPP, int x, int y, int Pitch)
{
	int Adj = (x & 1) << 1;
   	Ptr[0] = (uint8_t*)Planes[0] + ((x * BPP) >> 3) + y * Pitch;

	if (Format->Pixel.Flags & PF_YUV420)
	{
		Ptr[1] = (uint8_t*)Planes[1] + (x >> 1) + (y >> 1) * (Pitch >> 1);
		Ptr[2] = (uint8_t*)Planes[2] + (x >> 1) + (y >> 1) * (Pitch >> 1);
	}
	else
	if (Format->Pixel.Flags & PF_FOURCC)
		switch (Format->Pixel.FourCC)
		{
		case FOURCC_I420:
		case FOURCC_IYUV:
			Ptr[1] = (uint8_t*)Planes[0] + Format->Height * Pitch + (x >> 1) + (y >> 1) * (Pitch >> 1);
			Ptr[2] = Ptr[1] + ((Format->Height * Pitch) >> 2);
			break;
		case FOURCC_YV12: 
			Ptr[2] = (uint8_t*)Planes[0] + Format->Height * Pitch + (x >> 1) + (y >> 1) * (Pitch >> 1);
			Ptr[1] = Ptr[2] + ((Format->Height * Pitch) >> 2);
			break;
		case FOURCC_YUY2:
		case FOURCC_YUNV:
		case FOURCC_V422:
		case FOURCC_YUYV:
			Ptr[1] = Ptr[0]+1-Adj;
			Ptr[2] = Ptr[0]+3-Adj;
			break;
		case FOURCC_YVYU:
			Ptr[1] = Ptr[0]+3-Adj;
			Ptr[2] = Ptr[0]+1-Adj;
			break;
		case FOURCC_UYVY:
		case FOURCC_Y422:
		case FOURCC_UYNV:
			Ptr[1] = Ptr[0]-Adj;
			Ptr[2] = Ptr[0]+2-Adj;
			Ptr[0]++;
			break;
		case FOURCC_VYUY:
			Ptr[2] = Ptr[0]-Adj;
			Ptr[1] = Ptr[0]+2-Adj;
			Ptr[0]++;
			break;
		}
}

void BlitImage( void* Handle, const planes Dst, const planes Src, const planes SrcLast, int DstPitch, int SrcPitch )
{
	blitpack* Pack = (blitpack*)Handle;
	uint8_t* DstPtr[MAXPLANES];
	uint8_t* SrcPtr[MAXPLANES];
	bool_t OnlyDiff;
	int Width,Height;
	blit_soft* p;
	int Src2SrcLast;
	int DstStepX;
	int DstStepY;
	int DstX;
	int DstY;

	// nothing to do?
	if (!Pack || Pack->DstRect.Width<=0 || Pack->DstRect.Height<=0)
		return;

	OnlyDiff = (Pack->FX.Flags & BLITFX_ONLYDIFF) && SrcLast && SrcLast[0] != NULL;

	p = &Pack->Code[OnlyDiff];

	// calculate the Src and Dst pointers
	//   Src: always upperleft corner
	//   Dst: according to swapxy and mirroring

	Width = Pack->DstRect.Width;
	Height = Pack->DstRect.Height;

	if (p->SwapXY)
		SwapInt(&Width,&Height);

	if (DstPitch < 0)
		DstPitch = Pack->Dst.Pitch;
	if (SrcPitch < 0)
		SrcPitch = Pack->Src.Pitch;

	SurfacePtr(SrcPtr,Src,&Pack->Src,p->SrcBPP,Pack->SrcRect.x,Pack->SrcRect.y,SrcPitch);

	Src2SrcLast = 0;
	if (OnlyDiff)
		Src2SrcLast = (uint8_t*)SrcLast[0] - (uint8_t*)Src[0];

	DstStepX = p->DstBPP;
	DstStepY = DstPitch*8;
	DstX = Pack->DstRect.x;
	DstY = Pack->DstRect.y;

	if (p->FX.Direction & DIR_MIRRORLEFTRIGHT)
	{
		DstX += Pack->DstRect.Width-1;
		DstStepX = -DstStepX;
	}		

	if (p->FX.Direction & DIR_MIRRORUPDOWN)
	{
		DstY += Pack->DstRect.Height-1;
		DstStepY = -DstStepY;
	}

	if (p->SwapXY)
		SwapInt(&DstStepX,&DstStepY);

	SurfacePtr(DstPtr,Dst,&Pack->Dst,p->DstBPP,DstX,DstY,DstPitch);

	if (p->FX.Direction & DIR_MIRRORUPDOWN)
		DstPitch = -DstPitch;

	if (p->Slices)
	{
		const int DstBlock2 = 5;
		const int DstBlock = 32;

		int SrcBlock = (DstBlock * p->RScaleX) >> 4; //SrcBlock has to be even because of YUV
		int DstNext = (DstBlock * DstStepX) >> 3;
		int DstNextUV = PlanarYUV(&p->Dst) ? (DstNext >> (p->SwapXY?2:1)) : DstNext;
		int SrcNext = (SrcBlock * p->SrcBPP) >> 3;
		int SrcNextUV = PlanarYUV(&p->Src) ? (SrcNext >> 1) : SrcNext;

		if (Width > DstBlock && DstStepX > 0) // reverse order?
		{
			int Quot = Width >> DstBlock2;
			int Rem = Width & (DstBlock-1);

			DstPtr[0] += Quot*DstNext;
			DstPtr[1] += Quot*DstNextUV;
			DstPtr[2] += Quot*DstNextUV;

			SrcPtr[0] += Quot*SrcNext;
			SrcPtr[1] += Quot*SrcNextUV;
			SrcPtr[2] += Quot*SrcNextUV;

			if (Rem)
			{
				p->Entry(p,DstPtr,SrcPtr,DstPitch,SrcPitch,Rem,Height,Src2SrcLast);

				Width -= Rem;
			}

			DstNext = -DstNext;
			DstNextUV = -DstNextUV;
			SrcNext = -SrcNext;
			SrcNextUV = -SrcNextUV;

			DstPtr[0] += DstNext;
			DstPtr[1] += DstNextUV;
			DstPtr[2] += DstNextUV;

			SrcPtr[0] += SrcNext;
			SrcPtr[1] += SrcNextUV;
			SrcPtr[2] += SrcNextUV;
		}

		for (;Width > DstBlock;Width -= DstBlock)
		{
			p->Entry(p,DstPtr,SrcPtr,DstPitch,SrcPitch,DstBlock,Height,Src2SrcLast);

			DstPtr[0] += DstNext;
			DstPtr[1] += DstNextUV;
			DstPtr[2] += DstNextUV;

			SrcPtr[0] += SrcNext;
			SrcPtr[1] += SrcNextUV;
			SrcPtr[2] += SrcNextUV;
		}
	}

	p->Entry(p,DstPtr,SrcPtr,DstPitch,SrcPitch,Width,Height,Src2SrcLast); 
}

void BuildPalLookUp(blit_soft* p)
{
	//create a palette lookup with 3x3 bits RGB input

	int r,g,b,i;
	int Size = 1 << p->Dst.BitCount;
	uint8_t* LookUp = p->PalLookUp;
	for (r=0;r<256;r+=32)
		for (g=0;g<256;g+=32)
			for (b=0;b<256;b+=32)
			{
				int BestMatch = 0;
				int BestDiff = 0x7FFFFFFF;

				const rgb* q = p->DstPalette;
				for (i=0;i<Size;++i,++q)
				{
					int Diff = (q->r-r)*(q->r-r)+
							   (q->g-g)*(q->g-g)+
							   (q->b-b)*(q->b-b);

					if (Diff < BestDiff)
					{
						BestMatch = i;
						BestDiff = Diff;
					}
				}

				*(LookUp++) = (uint8_t)BestMatch;
			}
}

//!!! SH3 and MIPS just for testing
#if defined(_M_IX86) || defined(SH3) || defined(MIPS)

int UniversalType( const pixel* p, bool_t YUV )
{
	if (PlanarYUV(p)) return YUV ? 12:10;
	if (PackedYUV(p)) return YUV ? 13:11;

	if (p->Flags & PF_PALETTE)
	{
		if (p->BitCount==1) return 1;
		if (p->BitCount==2) return 2;
		if (p->BitCount==4) return 3;
		if (p->BitCount==8) return 4;
	}

	if (p->Flags & PF_RGB)
	{
		if (p->BitCount==8)  return 5;
		if (p->BitCount==16) return 6;
		if (p->BitCount==24) return 7;
		if (p->BitCount==32) return 8;
	}

	return -1;
}

void BlitUniversal(blit_soft* This, uint8_t** DstPtr,uint8_t** SrcPtr,int DstPitch,int SrcPitch,
	               int Width,int Height,int Src2SrcLast) 
{
	//this will be very-very slow, only for compability and testing

	int i,j,x,y;

	uint8_t* Src[3];
	uint8_t* Dst[3];

	int SrcType = This->SrcType;
	int DstType = This->DstType;

	int DstMask[3],DstPos[3];
	int SrcMask[3],SrcPos[3];
	int DitherMask[3];

	uint8_t* PalLookUp = This->PalLookUp;
	int Flags = This->FX.Flags;
	bool_t Dither = (Flags & BLITFX_DITHER) != 0 && !(This->Dst.Flags & PF_FOURCC);
	bool_t PalDither = Dither && (This->Dst.Flags & PF_PALETTE);

	const rgb* SrcPalette = This->SrcPalette;
	const rgb* DstPalette = This->DstPalette;

	int SrcStepX[16];
	int SrcStepY[16];
	int RScaleX = This->RScaleX;
	int RScaleY = This->RScaleY;

	int DstStepX = This->DstBPP;
	int DstStepX2 = This->DstBPP;
	int DstStepY = DstPitch << 3;
	int DstStepY2 = DstStepY >> 1;
	int dy = 0;
	int dy2 = 0;

	int R,G,B,Y,U,V;
	int sy0;
	int sy=0;
	int YAdd = This->FX.Brightness;

	uint32_t SrcInvert = 0;
	uint32_t DstInvert = 0;

	SrcMask[0] = This->Src.RBitMask;
	SrcMask[1] = This->Src.GBitMask;
	SrcMask[2] = This->Src.BBitMask;

	DstMask[0] = This->Dst.RBitMask;
	DstMask[1] = This->Dst.GBitMask;
	DstMask[2] = This->Dst.BBitMask;

	for (i=0;i<3;++i)
	{
		DitherMask[i] = Dither ? (1 << (8 - BitMaskSize(DstMask[i]))) - 1 : 0;
		SrcPos[i] = BitMaskPos(SrcMask[i]) + BitMaskSize(SrcMask[i]);
		DstPos[i] = BitMaskPos(DstMask[i]) + BitMaskSize(DstMask[i]);
		SrcMask[i] <<= 8;
		DstMask[i] <<= 8;
		Src[i] = SrcPtr[i];
		Dst[i] = DstPtr[i];
	}

	for (j=0;j<16;++j)
	{
		SrcStepX[j] = ((RScaleX * (j+1)) >> 4) - ((RScaleX * j) >> 4);
		SrcStepY[j] = ((RScaleY * (j+1)) >> 4) - ((RScaleY * j) >> 4);
	}

	if (This->FX.Direction & DIR_MIRRORLEFTRIGHT)
	{	
		DstStepX = -DstStepX;
		DstStepX2 = -DstStepX2;
		if (This->DstBPP < 8)
			dy = 8 - This->DstBPP;
	}

	if (This->SwapXY)
	{
		SwapInt(&DstStepX,&DstStepY);
		SwapInt(&DstStepX2,&DstStepY2);
	}

	R = (DitherMask[0] >> 1);
	G = (DitherMask[1] >> 1);
	B = (DitherMask[2] >> 1);
	Y=U=V=0;

	if (This->Src.Flags & PF_INVERTED)
		SrcInvert = (This->Src.BitCount>=32?0:(1 << This->Src.BitCount))-1;
	if (This->Dst.Flags & PF_INVERTED)
		DstInvert = (This->Dst.BitCount>=32?0:(1 << This->Dst.BitCount))-1;

	for (y=0;y<Height;++y)
	{
		int dx=dy;
		int dx2=dy2;
		int sx=0;

		for (x=0;x<Width;++x,dx+=DstStepX,sx+=SrcStepX[x&15])
		{
			uint8_t* q;
			const rgb* p;
			uint32_t v,w;

			switch (SrcType)
			{
			case 10: //Planar YUV->RGB
				R += ((Src[0][sx]+YAdd-16)*0x2568						        + 0x3343*(Src[2][sx>>1]-128)) /0x2000;
				G += ((Src[0][sx]+YAdd-16)*0x2568 - 0x0c92*(Src[1][sx>>1]-128)  - 0x1a1e*(Src[2][sx>>1]-128)) /0x2000;
				B += ((Src[0][sx]+YAdd-16)*0x2568 + 0x40cf*(Src[1][sx>>1]-128))                               /0x2000;

				SAT(R)
				SAT(G)
				SAT(B)
				break;
			case 11: //Packed YUV->RGB
				R += ((Src[0][sx*2]+YAdd-16)*0x2568						              + 0x3343*(Src[2][4*(sx>>1)]-128)) /0x2000;
				G += ((Src[0][sx*2]+YAdd-16)*0x2568 - 0x0c92*(Src[1][4*(sx>>1)]-128)  - 0x1a1e*(Src[2][4*(sx>>1)]-128)) /0x2000;
				B += ((Src[0][sx*2]+YAdd-16)*0x2568 + 0x40cf*(Src[1][4*(sx>>1)]-128))                                   /0x2000;

				SAT(R)
				SAT(G)
				SAT(B)
				break;
			case 12: //Planar YUV->YUV
				Y = Src[0][sx]+YAdd;
				U = Src[1][sx>>1];
				V = Src[2][sx>>1];
				SAT(Y)
				break;
			case 13: //Packed YUV->YUV
				Y = Src[0][sx*2]+YAdd;
				U = Src[1][4*(sx>>1)];
				V = Src[2][4*(sx>>1)];
				SAT(Y)
				break;
			case 1: //Pal1->RGB
				p = &SrcPalette[ ((Src[0][sx>>3] >> ((~sx)&7)) & 1) ^ SrcInvert];
				R += p->r; G += p->g; B += p->b;
				break;
			case 2: //Pal2->RGB
				p = &SrcPalette[ ((Src[0][sx>>2] >> (((~sx)&3)*2)) & 3) ^ SrcInvert];
				R += p->r; G += p->g; B += p->b;
				break;
			case 3: //Pal4->RGB
				p = &SrcPalette[ ((Src[0][sx>>1] >> (((~sx)&1)*4)) & 15) ^ SrcInvert];
				R += p->r; G += p->g; B += p->b;
				break;
			case 4: //Pal8->RGB
				p = &SrcPalette[Src[0][sx] ^ SrcInvert];
				R += p->r; G += p->g; B += p->b;
				break;
			case 5: //RGB8->RGB
				v = Src[0][sx];
				v ^= SrcInvert;
				v <<= 8;
				R += (v & SrcMask[0]) >> SrcPos[0];
				G += (v & SrcMask[1]) >> SrcPos[1];
				B += (v & SrcMask[2]) >> SrcPos[2];
				break;
			case 6: //RGB16->RGB
				v = ((uint16_t*)Src[0])[sx];
				v ^= SrcInvert;
				v <<= 8;
				R += (v & SrcMask[0]) >> SrcPos[0];
				G += (v & SrcMask[1]) >> SrcPos[1];
				B += (v & SrcMask[2]) >> SrcPos[2];
				break;
			case 7: //RGB24->RGB
				v = Src[0][sx*3] | (Src[0][sx*3+1] << 8) | (Src[0][sx*3+2] << 16);
				v ^= SrcInvert;
				v <<= 8;
				R += (v & SrcMask[0]) >> SrcPos[0];
				G += (v & SrcMask[1]) >> SrcPos[1];
				B += (v & SrcMask[2]) >> SrcPos[2];
				break;
			default: //RGB32->RGB
				v = ((uint32_t*)Src[0])[sx];
				v ^= SrcInvert;
				v <<= 8;
				R += (v & SrcMask[0]) >> SrcPos[0];
				G += (v & SrcMask[1]) >> SrcPos[1];
				B += (v & SrcMask[2]) >> SrcPos[2];
				break;
			}

			q = Dst[0]+(dx >> 3);
			switch (DstType)
			{
			case 10: //RGB->Planar YUV
				Y = ( (2105 * R) + (4128 * G) + ( 802 * B))/0x2000 + 16;
				V = ( (3596 * R) - (3015 * G) - ( 582 * B))/0x2000 + 128;
				U = (-(1212 * R) - (2384 * G) + (3596 * B))/0x2000 + 128;

				*q = (uint8_t)Y;
				Dst[1][dx2 >> 3] = (uint8_t)U;
				Dst[2][dx2 >> 3] = (uint8_t)V;
				if (x & 1)
					dx2+=DstStepX2;
				R=G=B=0;
				break;
			case 11: //RGB->Packed YUV
				Y = ( (2105 * R) + (4128 * G) + ( 802 * B))/0x2000 + 16;
				V = ( (3596 * R) - (3015 * G) - ( 582 * B))/0x2000 + 128;
				U = (-(1212 * R) - (2384 * G) + (3596 * B))/0x2000 + 128;
				SAT(Y)
				SAT(U)
				SAT(V)
				*q = (uint8_t)Y;
				Dst[1][4*(dx >> 5)] = (uint8_t)U;
				Dst[2][4*(dx >> 5)] = (uint8_t)V;
				R=G=B=0;
				break;
			case 12: //YUV->Planar YUV
				*q = (uint8_t)Y;
				Dst[1][dx2 >> 3] = (uint8_t)U;
				Dst[2][dx2 >> 3] = (uint8_t)V;
				if (x & 1)
					dx2+=DstStepX2;
				break;
			case 13: //YUV->Packed YUV
				*q = (uint8_t)Y;
				Dst[1][4*(dx >> 5)] = (uint8_t)U;
				Dst[2][4*(dx >> 5)] = (uint8_t)V;
				break;
			case 1: //RGB->Pal1
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				w = PalLookUp[w >> 8];
				w ^= DstInvert;
				*q &= ~(1 << ((~dx)&7));
				*q |=  (w << ((~dx)&7));

				if (PalDither)
				{
					p = &DstPalette[w ^ DstInvert];
					R -= p->r; G -= p->g; B -= p->b;
					SAT(R)
					SAT(G)
					SAT(B)
				}
				else
					R=G=B=0;
				break;
			case 2: //RGB->Pal2
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				w = PalLookUp[w >> 8];
				w ^= DstInvert;
				*q &= ~(3 << ((~dx)&6));
				*q |=  (w << ((~dx)&6));

				if (PalDither)
				{
					p = &DstPalette[w ^ DstInvert];
					R -= p->r; G -= p->g; B -= p->b;
					SAT(R)
					SAT(G)
					SAT(B)
				}
				else
					R=G=B=0;
				break;
			case 3: //RGB->Pal4
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				w = PalLookUp[w >> 8];
				w ^= DstInvert;
				*q &= ~(15 << ((~dx)&4));
				*q |=  (w  << ((~dx)&4));

				if (PalDither)
				{
					p = &DstPalette[w ^ DstInvert];
					R -= p->r; G -= p->g; B -= p->b;
					SAT(R)
					SAT(G)
					SAT(B)
				}
				else
					R=G=B=0;
				break;
			case 4: //RGB->Pal8
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				w = PalLookUp[w >> 8];
				w ^= DstInvert;
				*q = (uint8_t)w;
				if (PalDither)
				{
					p = &DstPalette[w ^ DstInvert];
					R -= p->r; G -= p->g; B -= p->b;
					SAT(R)
					SAT(G)
					SAT(B)
				}
				else
					R=G=B=0;
				break;
			case 5: //RGB->RGB8
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				*q = (uint8_t)((w >> 8) ^ DstInvert);
				R &= DitherMask[0];
				G &= DitherMask[1];
				B &= DitherMask[2];
				break;
			case 6: //RGB->RGB16
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				*(uint16_t*)q = (uint16_t)((w >> 8) ^ DstInvert);
				R &= DitherMask[0];
				G &= DitherMask[1];
				B &= DitherMask[2];
				break;
			case 7: //RGB->RGB24
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				w >>= 8;
				w ^= DstInvert;
				q[0] = (uint8_t)(w);
				q[1] = (uint8_t)(w >> 8);
				q[2] = (uint8_t)(w >> 16);
				R &= DitherMask[0];
				G &= DitherMask[1];
				B &= DitherMask[2];
				break;
			default: //RGB->RGB32
				w = ((R << DstPos[0]) & DstMask[0]) |
					((G << DstPos[1]) & DstMask[1]) |
					((B << DstPos[2]) & DstMask[2]);
				*(uint32_t*)q = (w >> 8) ^ DstInvert;
				R &= DitherMask[0];
				G &= DitherMask[1];
				B &= DitherMask[2];
				break;
			}
		}

		dy += DstStepY;
		if (y & 1)
			dy2 += DstStepY2;

		sy0 = sy;
		sy += SrcStepY[y&15];
		Src[0] += SrcPitch * (sy - sy0);
		Src[1] += (SrcPitch >> 1) * ((sy >> 1) - (sy0 >> 1));
		Src[2] += (SrcPitch >> 1) * ((sy >> 1) - (sy0 >> 1));
	}
}

#endif

bool_t BlitCompile(blit_soft* p,const video* NewDst,const video* NewSrc,
		  		   const blitfx* NewFX,bool_t NewOnlyDiff)
{
	int i;
	bool_t VMemRotated,Gray;

	if (EqPixel(&p->Dst,&NewDst->Pixel) && 
		EqPixel(&p->Src,&NewSrc->Pixel) && 
		EqBlitFX(&p->FX,NewFX) &&
		NewOnlyDiff == p->OnlyDiff)
		return p->Entry != NULL;

	CodeStart(&p->Code);

	p->Dst = NewDst->Pixel;
	p->Src = NewSrc->Pixel;
	p->FX = *NewFX;
	p->OnlyDiff = NewOnlyDiff;
	p->SwapXY = (p->FX.Direction & DIR_SWAPXY) != 0;
	VMemRotated = (p->FX.Flags & BLITFX_VMEMROTATED) != 0;
	p->Slices = (p->FX.Flags & BLITFX_AVOIDTEARING) && (p->SwapXY != VMemRotated);

	Gray = (p->Dst.Flags & PF_PALETTE) && (p->Dst.BitCount == 4 || p->Dst.BitCount == 2);
	p->RScaleX = CalcRScale(p->FX.ScaleX,Gray);
	p->RScaleY = CalcRScale(p->FX.ScaleY,Gray);

	p->SrcDoubleX = p->RScaleX == 8 ? 1:0;
	p->SrcDoubleY = p->RScaleY == 8 ? 1:0;
	p->DstDoubleX = p->SrcDoubleX;
	p->DstDoubleY = p->SrcDoubleY;
	if (p->SwapXY)
		SwapInt(&p->DstDoubleX,&p->DstDoubleY);

	p->SrcBPP = GetBPP(&p->Src);
	p->DstBPP = GetBPP(&p->Dst);
	p->DstBPP2 = -3;
	for (i=p->DstBPP;i>1;i>>=1)
		++p->DstBPP2;

	p->SrcPalette = DefaultPal(&p->Src);
	p->DstPalette = DefaultPal(&p->Dst);

	if (p->DstPalette)
	{
		p->Dst.RBitMask = 0x1C0;
		p->Dst.GBitMask = 0x038;
		p->Dst.BBitMask = 0x007;
		BuildPalLookUp(p);
	}

	p->DirX = (p->FX.Direction & DIR_MIRRORLEFTRIGHT) ? -1:1;
	p->RSize = BitMaskSize(p->Dst.RBitMask);
	p->GSize = BitMaskSize(p->Dst.GBitMask);
	p->BSize = BitMaskSize(p->Dst.BBitMask);
	p->RPos = BitMaskPos(p->Dst.RBitMask);
	p->GPos = BitMaskPos(p->Dst.GBitMask);
	p->BPos = BitMaskPos(p->Dst.BBitMask);

#if defined(ARM) || defined(SH3) || defined(MIPS)
	if (PlanarYUV(&p->Src))
	{
#if defined(ARM)
		if (PlanarYUV(&p->Dst))
			Fix_YUV_YUV(p);
#endif
		if ((p->Dst.Flags & (PF_RGB|PF_PALETTE)) && (p->Dst.BitCount == 8 || p->Dst.BitCount == 16))
		{
#if defined(ARM)
			if (p->RScaleX == 16 && p->RScaleY == 16)
				Fix_RGB_UV(p);
			else
			if ((p->RScaleX == 8 || p->RScaleX == 16) && 
				(p->RScaleY == 8 || p->RScaleY == 16) && !(p->FX.Flags & (BLITFX_DITHER|BLITFX_ARITHSTRETCH)))
				Fix_RGB_UV(p);
			else
//			if (RScaleX == 8 && RScaleY == 8) 
//				Double_RGB_UV(p);
//			else
			if (p->RScaleX == 32 && p->RScaleY == 32)
				Half_RGB_UV(p);
			else
				Stretch_RGB_UV(p);
#else
			Universal_RGB_UV(p);
#endif
		}
		else
		if (Gray)
			Fix_Gray_UV(p);
	}
#endif

	CodeBuild(&p->Code);
	if (p->Code.Size)
		p->Entry = (blitsoftentry)p->Code.Code;
	else
		p->Entry = NULL;

//!!! SH3 and MIPS just for testing
#if defined(_M_IX86) || defined(SH3) || defined(MIPS)
	if (!p->Entry)
	{
		// use YUV internal calulaction?
		bool_t YUV = AnyYUV(&p->Dst) && AnyYUV(&p->Src);

		p->DstType = UniversalType(&p->Dst,YUV);
		p->SrcType = UniversalType(&p->Src,YUV);

		if (p->SrcType>=0 && p->DstType>=0)
			p->Entry = BlitUniversal; 
	}
#endif

	return p->Entry != NULL;
}

int64_t CMul( blit_soft* p, int64_t v, bool_t UV )
{
	int m;

	if (UV)
	{
		m = p->FX.Saturation;
		if (m<0) m >>= 1; // adjust negtive interval: -128..0 -> -64..0
		m = 4*m+256;
		v = (v*m) >> 8;
	}

	m = p->FX.Contrast;
	if (m<0) m >>= 1; // adjust negtive interval: -128..0 -> -64..0
	m = 4*m + 256;
	v = (v*m) >> 8;

	return v;
}

int64_t CAdd( blit_soft* p, int64_t v, int v0, int vUV )
{
	int m;

	m = p->FX.Saturation;
	if (m<0) m >>= 1; // adjust negtive interval: -128..0 -> -64..0
	m = 4*m+256;
	v += ((256 - m) * vUV) >> 1;

	m = p->FX.Contrast;
	if (m<0) m >>= 1; // adjust negtive interval: -128..0 -> -64..0
	m = 4*m + 256;
	v += ((256 - m)*(v0 - v)) >> 8;

	return v;
}

INLINE bool_t InRange32( int64_t v )
{
	return v >= -MAX_INT && v <= MAX_INT;
}

void CalcColor( blit_soft* p )
{
	if (p->QAdd)
	{
		//saturation: signed 32bit / 3 (qdadd will triple)
		int64_t YMul = CMul(p,0x63C000,0);
		int64_t RVMul = CMul(p,0x88B2AA,1);
		int64_t RAdd = CAdd(p,p->FX.Brightness*0x63C000 - (int64_t)0x75400000,0x2AAAAAAA,0x88B2AA);
		int64_t GUMul = CMul(p,-0x218555,1);
		int64_t GVMul = CMul(p,-0x45A555,1);
		int64_t GAdd = CAdd(p,p->FX.Brightness*0x63C000 + (int64_t)0x02AEAAAA,0x2AAAAAAA,-0x218555-0x45A555);
		int64_t BUMul = CMul(p,0xACD2AA,1);
		int64_t BAdd = CAdd(p,p->FX.Brightness*0x63C000 - (int64_t)0x87500000,0x2AAAAAAA,0xACD2AA);

		if (InRange32(YMul*0xF0) && 
			InRange32(RVMul*0x10+RAdd) &&
			InRange32(RVMul*0xF0+RAdd) &&
			InRange32((GVMul+GUMul)*0x10+GAdd) &&
			InRange32((GVMul+GUMul)*0xF0+GAdd) &&
			InRange32(BUMul*0x10+BAdd) &&
			InRange32(BUMul*0xF0+BAdd))
		{
			p->_YMul = (int)YMul;
			p->_RVMul = (int)RVMul;
			p->_RAdd = (int)RAdd;
			p->_GUMul = (int)GUMul;
			p->_GVMul = (int)GVMul;
			p->_GAdd = (int)GAdd;
			p->_BUMul = (int)BUMul;
			p->_BAdd = (int)BAdd;
		}
		else
			p->QAdd = 0;
	}

	if (!p->QAdd)
	{
		//saturation: unsigned 24bit
		p->_YMul = (int)CMul(p,0x12B40,0);
		p->_RVMul = (int)CMul(p,0x19A18,1);
		p->_RAdd = (int)CAdd(p,p->FX.Brightness*0x12B40 - 0x0DFC000,0x800000,0x19A18);
		p->_GUMul = (int)CMul(p,-0x06490,1);
		p->_GVMul = (int)CMul(p,-0x0D0F0,1);
		p->_GAdd = (int)CAdd(p,p->FX.Brightness*0x12B40 + 0x0880C00,0x800000,-0x06490-0x0D0F0);
		p->_BUMul = (int)CMul(p,0x20678,1);
		p->_BAdd = (int)CAdd(p,p->FX.Brightness*0x12B40 - 0x115F000,0x800000,0x20678);
	}
}
