/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#ifndef __NODE_H
#define __NODE_H

#define MAXNODE	256
#define MAXDATA	(256*sizeof(tchar_t))

// buffer full: data is not processed, retry later
// need more data: data is processed, but need more to continue

//----------------------------------------------------------------
// data types

// bool_t
#define TYPE_BOOL			1
// int (format: optional tchar_t[] enumerate or min..max)
#define TYPE_INT			2
// fraction_t
#define TYPE_FRACTION		3
// null terminated tchar_t[]
#define TYPE_STRING			4
// rect
#define TYPE_RECT			5
// point
#define TYPE_POINT			6
// rgb_t
#define TYPE_RGB			7
// equalizer
#define TYPE_EQUALIZER		8
// node* (format1: node class)
#define TYPE_NODE			10
// output pin (format1: packet type)
#define TYPE_PIN			11
// input packet (format1: packet type)
#define TYPE_PACKET			12
// tick_t
#define TYPE_TICK			13
// blitfx
#define TYPE_BLITFX			14
// indicates that a node is part of a class
#define TYPE_CLASS			15
// virtual key code (int)
#define TYPE_HOTKEY			16
// settings func
#define TYPE_FUNC			17

//----------------------------------------------------------------
// data flags

#define DF_RDONLY			0x0001
#define DF_SETUP			0x0002
#define DF_HIDDEN			0x0004
#define DF_MINMAX			0x0008
#define DF_ENUM				0x0010
#define DF_INPUT			0x0020
#define DF_OUTPUT			0x0040
#define DF_OWNREG			0x0080
#define DF_SOFTRESET		0x0100
#define DF_RESTART			0x0200
#define DF_CHECKLIST		0x0400
#define DF_HEX				0x0800
#define DF_ENUMCLASS		0x1000
#define DF_PERCENT			0x2000
#define DF_NOWRAP			0x4000
#define DF_GAP				0x8000
#define DF_RESYNC			0x10000
#define DF_MSEC				0x20000
#define DF_KBYTE			0x40000
#define DF_NEG				0x80000

#define PERCENT_ONE			1024

typedef struct datadef
{
	int	No;
	int	Type;
	int Flags;
	int Format1; // min, enumdef, node class
	int	Format2; // max,

	const tchar_t* Name;	// filled by enum
	int Class;				// filled by enum
	int Size;				// filled by enum

} datadef;

#define DATADEF_END { -1 }

typedef struct enumdef
{
	int	Class;
	int	No;
} enumdef;

#define ENUMDEF_END { -1 }

//---------------------------------------------------------------
// node

#define NODE_CLASS			FOURCC('N','O','D','E')
// unique fourcc id of this node (int)
#define NODE_ID				1
// test node class (if setting one fourcc code returns ERR_NONE) (int)
#define NODE_TYPE			2

extern DLL const datadef NodeParams[];

// enumerate parameters
typedef	int (*nodeenum)(void* This,int EnumNo,datadef* Out);
// get parameter
typedef	int (*nodeget)(void* This,int No,void* Data,int Size);
// set parameter or send data
typedef	int (*nodeset)(void* This,int No,const void* Data,int Size);

typedef struct node
{
	nodeenum	Enum;
	nodeget		Get;	
	nodeset		Set;

} node;

//---------------------------------------------------------------
// input pin data type

// special flag for get/set data format structure
#define PACKET_FORMAT	0x10000

// format structure: video 
#define PACKET_VIDEO	1
// format structure: audio
#define PACKET_AUDIO	2
// format structure: void
#define PACKET_COMMENT	3

typedef struct packet
{
	planes Data;
	planes LastData;	// optional (for video it may increase performance)
	int Length;			// length of data (in bytes, per planes, not filled for surfaces)
	tick_t RefTime;		// packet time (-1 if unknown)
	tick_t CurrTime;	// current time (-1 for forcing decoding)
	int DropLevel;		// try to save time (1:don't display frame 2:skip until next keyframe)

} packet;

//---------------------------------------------------------------
// output pin data type
typedef struct pin
{
	node* Node;
	int No;

} pin;

//---------------------------------------------------------------

// create dynamic node
typedef	void* (*dynnodecreate)(void* This);
// delete dynamic node
typedef	void (*dynnodedelete)(void* This);

//---------------------------------------------------------------
// functions managing nodes

#define PRI_MINIMUM			1
#define PRI_DEFAULT		 1000
#define PRI_MAXIMUM		10000

DLL int NodeRegister( node*, int Priority ); //loads setup too
DLL int NodeUnRegister( node* );
DLL node* NodeEnum( int Class, int No );
DLL node* NodeEnumInside( int Class, int No, node* Skip );
DLL node* NodeFind( int ById );
DLL bool_t NodeDataDef( node* p, int No, datadef* Out );
DLL bool_t NodeIsClass( node* p, int Class );
DLL bool_t NodeInsideIsClass();

DLL bool_t NodeLoadValue( int Id, void* Data, int Size, int* Type );
DLL void NodeSaveValue( int Id, const void* Data, int Size, int Type );
DLL void NodeLoadSetup( node* );
DLL void NodeSaveSetup( node* );

// DLL handling
typedef void (STDCALL *funcany)();

DLL void Node_Init( const tchar_t* ProgamId, const tchar_t* Version, funcany Any, const tchar_t* CmdLine );
DLL void Node_Done();

DLL void NodePlugins();
DLL const tchar_t* NodeProgramId();
DLL const tchar_t* NodeVersion();
DLL const tchar_t* NodeCmdLine();

DLL bool_t NodeLocate( void* Ptr, tchar_t* OutName, int* OutBase );
DLL void NodeDump( FILE* );

DLL void NodeError( node* Node, int Class, int No, ... );
DLL pin NodeErrorPin;

// functions which should be exported in node DLLs
DLLEXPORT int DLLRegister();
DLLEXPORT void DLLUnRegister();

#endif
