/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "stdafx.h"

#if defined(_WIN32)

#include <windows.h>

#ifndef ERROR_DEVICE_REMOVED
#define ERROR_DEVICE_REMOVED             1617L
#endif

typedef struct file
{
	stream VMT;

	tchar_t URL[MAXPATH];
	HANDLE Handle;
	int Length;
	int Pos;
	bool_t Silent;

	HANDLE Find;
	const tchar_t* Exts;
	bool_t ExtFilter;
	WIN32_FIND_DATA FindData;

} file;

static int EnumType( void* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,StreamParams,NULL);
}

static int Get( file* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(FILE_ID,int); break;
	case STREAM_MIME: GETSTRING(T("file")); break;
	case STREAM_URL: GETSTRING(p->URL); break;
	case STREAM_SILENT: GETVALUE(p->Silent,bool_t); break;
	case STREAM_LENGTH: GETVALUECOND(p->Length,int,p->Length>=0); break;
	}
	return Result;
}

static int Open( file* p, const tchar_t* URL, bool_t ReOpen )
{
	if (p->Handle)
		CloseHandle(p->Handle);
	p->Handle = 0;
	p->Length = -1;
	if (!ReOpen)
		p->URL[0] = 0;
	else
		Sleep(200);

	if (URL && URL[0])
	{
		HANDLE Handle = CreateFile(URL,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
		if (Handle == INVALID_HANDLE_VALUE)
		{
			if (!ReOpen && !p->Silent)
				NodeError(NULL,ERR_ID,ERR_FILE_NOT_FOUND,URL);
			return ERR_FILE_NOT_FOUND;
		}

		p->Handle = Handle;
		tcsncpy(p->URL,URL,sizeof(p->URL)/sizeof(tchar_t));
		p->URL[MAXPATH-1]=0;

		p->Length = GetFileSize(Handle,NULL);

		if (ReOpen)
			p->Pos = SetFilePointer(p->Handle,p->Pos,NULL,FILE_BEGIN);
		else
			p->Pos = 0;
	}
	else
		p->Length = -1;
	return ERR_NONE;
}

static int Set( file* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case STREAM_SILENT: SETVALUE(p->Silent,bool_t,ERR_NONE); break;
	case STREAM_URL:
		Result = Open(p,(const tchar_t*)Data,0);
		break;
	}
	return Result;
}

static int Read(file* p,void* Data,int Size)
{
	unsigned long Readed;
	DWORD Error;

	//DEBUG_MSG2("FileRead: %08x %d",SetFilePointer(p->Handle,0,NULL,FILE_CURRENT),Size);

	if (ReadFile(p->Handle,Data,Size,&Readed,NULL))
	{
		DEBUG_MSG2(T("READ pos:%d len:%d"),p->Pos,Readed);
		p->Pos += Readed;
		return Readed;
	}

	Error = GetLastError();

	DEBUG_MSG2(T("READ pos:%d error:%d"),p->Pos,Error);

	if (Error == ERROR_DEVICE_REMOVED || Error == ERROR_INVALID_HANDLE)
		Open(p,p->URL,1);

	return -1;
}

static int Seek(file* p,int Pos,int SeekMode)
{
	int ReTry=3;
	int Result;
	DWORD Error;

	do
	{
		//DEBUG_MSG3("FileSeek: %08x %08x %d",SetFilePointer(p->Handle,0,NULL,FILE_CURRENT),Pos,SeekMode);

		switch (SeekMode)
		{
		default:
		case SEEK_SET: SeekMode = FILE_BEGIN; break;
		case SEEK_CUR: SeekMode = FILE_CURRENT; break;
		case SEEK_END: SeekMode = FILE_END; break;
		}

		Result = SetFilePointer(p->Handle,Pos,NULL,SeekMode);

		if (Result != -1)
		{
			p->Pos = Result;
			break;
		}

		Error = GetLastError();
		if (Error != ERROR_DEVICE_REMOVED && Error != ERROR_INVALID_HANDLE)
			break;

		Open(p,p->URL,1);
	}
	while (--ReTry>0);

	return Result;
}

static int EnumDir(file* p,const tchar_t* URL,const tchar_t* Exts,bool_t ExtFilter,streamdir* Item)
{
	if (URL)
	{
		DWORD Attrib;
		tchar_t Path[MAXPATH];

		p->Exts = Exts;
		p->ExtFilter = ExtFilter;

		if (p->Find != INVALID_HANDLE_VALUE)
		{
			FindClose(p->Find);
			p->Find = INVALID_HANDLE_VALUE;
		}

		stprintf(Path,T("%s\\*.*"),URL);

		Attrib = GetFileAttributes(URL);
		if (Attrib != (DWORD)-1 && !(Attrib & FILE_ATTRIBUTE_DIRECTORY))
			return ERR_NOT_DIRECTORY;

		p->Find = FindFirstFile( Path, &p->FindData );
	}

	Item->FileName[0] = 0;

	while (!Item->FileName[0] && p->Find != INVALID_HANDLE_VALUE)
	{
		if (tcscmp(p->FindData.cFileName,T("."))!=0 &&	tcscmp(p->FindData.cFileName,T(".."))!=0)
		{
			tcscpy(Item->FileName,p->FindData.cFileName);

			if (p->FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				Item->Size = -1;
			else
			{
				Item->Size = p->FindData.nFileSizeLow;
				Item->Type = CheckExts(Item->FileName,p->Exts);

				if (!Item->Type && p->ExtFilter)
					Item->FileName[0] = 0; // skip
			}
		}
			
		if (!FindNextFile(p->Find,&p->FindData))
		{
			FindClose(p->Find);
			p->Find = INVALID_HANDLE_VALUE;
		}
	}

	return Item->FileName[0] ? ERR_NONE : ERR_END_OF_FILE;
}

static void* Create( file* p )
{
	file* i = (file*) Alloc(sizeof(file));
	*i = *p;
	return i;
}

static void Delete( file* p )
{
	Open(p,NULL,0);
	if (p->Find != INVALID_HANDLE_VALUE)
		FindClose(p->Find);
	Free(p);
}

static file File = 
{
	EnumType,
	Get,
	Set,
	Create,
	Delete,

	Read,
	Seek,
	EnumDir,
};

void File_Init()
{
	File.Find = INVALID_HANDLE_VALUE;

	NodeRegister((node*)&File,PRI_MINIMUM);
}

void File_Done()
{
	NodeUnRegister((node*)&File);
}

#endif
