/*
 * Copyright (c) 2004 Mike Matsnev.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Absolutely no warranty of function or purpose is made by the author
 *    Mike Matsnev.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: unixio.c,v 1.1.1.1 2004/04/27 13:04:47 mike Exp $
 * 
 */

#include <sys/mman.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "Parser.h"

struct UnixInputFile {
  struct InputFile    base;
  int		      fd;
  char		      errmsg[128];
};

typedef struct UnixInputFile UnixInputFile;

static int	  uif_read(InputFile *inf, void *buffer, int count) {
  UnixInputFile	*uif = (UnixInputFile *)inf;
  int		rd = read(uif->fd, buffer, count);

  if (rd < 0) {
    strlcpy(uif->errmsg, strerror(errno), sizeof(uif->errmsg));
    return -1;
  }

  return rd;
}

static long long  uif_seek(InputFile *inf, long long where, int how) {
  UnixInputFile	*uif = (UnixInputFile *)inf;
  off_t		newpos = lseek(uif->fd, where, how);

  if (newpos < 0) {
    strlcpy(uif->errmsg, strerror(errno), sizeof(uif->errmsg));
    return -1;
  }

  return newpos;
}

static void uif_close(InputFile *inf) {
  UnixInputFile	*uif = (UnixInputFile *)inf;

  close(uif->fd);
  free(uif);
}

static void *uif_memalloc(InputFile *inf, unsigned size) {
  UnixInputFile	*uif = (UnixInputFile *)inf;
  void	*mem = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_ANON, -1, 0);
  if (mem == MAP_FAILED) {
    strlcpy(uif->errmsg, strerror(errno), sizeof(uif->errmsg));
    return NULL;
  }

  return mem;
}

static void uif_memfree(InputFile *inf, void *mem, unsigned size) {
  munmap(mem, size);
}

static UnixInputFile  *uif_open(const char *filename)
{
  UnixInputFile	*uif;

  uif = malloc(sizeof(*uif));
  if (uif == NULL)
    return NULL;

  uif->fd = open(filename, O_RDONLY|O_DIRECT);
  if (uif->fd < 0) {
    free(uif);
    return NULL;
  }

  uif->base.last_error = uif->errmsg;
  uif->base.read = uif_read;
  uif->base.seek = uif_seek;
  uif->base.close = uif_close;
  uif->base.memalloc = uif_memalloc;
  uif->base.memfree = uif_memfree;

  return uif;
}

InputFile *fileio_open(/* in */	  const char *filename,
		       /* out */  char *errormsg,
		       /* in */	  int msglen)
{
  InputFile *inf = (InputFile *)uif_open(filename);
  if (inf == NULL) {
    strlcpy(errormsg, strerror(errno), sizeof(errormsg));
    return NULL;
  }
  
  return inf;
}

