/*
 * Copyright (c) 2004 Mike Matsnev.  All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice immediately at the beginning of the file, without modification,
 *    this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Absolutely no warranty of function or purpose is made by the author
 *    Mike Matsnev.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: win32io.c,v 1.2 2004/04/28 14:30:42 mike Exp $
 * 
 */

#include <windows.h>
#include <tchar.h>

#include "Parser.h"

// Win32 based InputFile
#ifndef INVALID_SET_FILE_POINTER
#define INVALID_SET_FILE_POINTER 0xffffffff
#endif

#ifdef _WIN32_WCE
#define	OPENFLAGS 0
#else
#define	OPENFLAGS FILE_FLAG_NO_BUFFERING
#endif

struct Win32InputFile {
  struct InputFile  base;
  HANDLE	    fh;
  char		    errmsg[128];
};

typedef struct Win32InputFile Win32InputFile;

static void	LastErrorMsg(char *buffer, int bufsize) {
#ifdef _WIN32_WCE
  TCHAR   tmp[128];
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,NULL,GetLastError(),0,tmp,sizeof(tmp)/sizeof(tmp[0]),NULL);
  WideCharToMultiByte(CP_ACP,0,tmp,-1,buffer,bufsize,NULL,NULL);
#else
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM,NULL,GetLastError(),0,buffer,bufsize,NULL);
#endif
}


static int	Win32Read(InputFile *inf,void *buffer,int count) {
  Win32InputFile  *sif = (Win32InputFile *)inf;
  DWORD		  nread;

  if (!ReadFile(sif->fh,buffer,count,&nread,NULL)) {
    LastErrorMsg(sif->errmsg,sizeof(sif->errmsg));
    return -1;
  }

  return nread;
}

static longlong  Win32Seek(InputFile *inf,longlong where,int how) {
  Win32InputFile  *sif = (Win32InputFile *)inf;

  LARGE_INTEGER	  li;

  li.QuadPart = where;
  li.LowPart=SetFilePointer(sif->fh,li.LowPart,&li.HighPart,how);
  if (li.LowPart==INVALID_SET_FILE_POINTER && GetLastError()!=NO_ERROR) {
    LastErrorMsg(sif->errmsg,sizeof(sif->errmsg));
    return -1;
  }

  return li.QuadPart;
}

static void	  Win32Close(InputFile *inf) {
  Win32InputFile  *sif = (Win32InputFile *)inf;

  CloseHandle(sif->fh);
  free(sif);
}

static void *Win32Alloc(InputFile *inf, unsigned size) {
  return VirtualAlloc(NULL,size,MEM_RESERVE|MEM_COMMIT,PAGE_READWRITE);
}

static void Win32Free(InputFile *inf, void *ptr, unsigned size) {
  VirtualFree(ptr, size, MEM_DECOMMIT|MEM_RELEASE);
}

static Win32InputFile	*Win32Open(const char *fname)
{
  Win32InputFile  *sif;
#ifdef _WIN32_WCE
  TCHAR		  tmp[MAX_PATH];

  MultiByteToWideChar(CP_ACP,0,fname,-1,tmp,sizeof(tmp)/sizeof(tmp[0]));
#define fname tmp
#endif

  sif = malloc(sizeof(*sif));
  if (sif == NULL) {
    SetLastError(ERROR_OUTOFMEMORY);
    return NULL;
  }

  sif->fh = CreateFile(fname,GENERIC_READ,FILE_SHARE_READ,NULL,OPEN_EXISTING,OPENFLAGS,NULL);
  if (sif->fh == INVALID_HANDLE_VALUE) {
    free(sif);
    return NULL;
  }

#undef fname

  sif->base.last_error = sif->errmsg;
  sif->base.read = Win32Read;
  sif->base.seek = Win32Seek;
  sif->base.close = Win32Close;
  sif->base.memalloc = Win32Alloc;
  sif->base.memfree = Win32Free;

  return sif;
}

InputFile *fileio_open(/* in */	  const char *filename,
		       /* out */  char *errormsg,
		       /* in */	  int msglen)
{
  InputFile *sif = (InputFile *)Win32Open(filename);
  if (sif == NULL) {
    LastErrorMsg(errormsg,msglen);
    return NULL;
  }

  return sif;
}
