/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "..\common\common.h"
#include "mpegaudio.h"
#include "libid3tag\id3tag.h"
#include "libmad\frame.h"
#include "libmad\synth.h"
#include <math.h>

#define LIBMAD_INPUT		0x100
#define LIBMAD_OUTPUT		0x101
#define LIBMAD_EQ			0x102
#define LIBMAD_COMMENTS		0x103

typedef struct libmad
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;
	pin Comments;

	equalizer Eq;

	merge Merge;
	struct mad_stream Stream;
	struct mad_frame Frame;
	struct mad_synth Synth;

	// output data
	packet Packet;
	tick_t RefTime;
	bool_t Skip; // skip next frame
	bool_t Pending;
	int FormatSet;
	int Total;

	mad_fixed_t Scale;
	mad_fixed_t EqTable[32];

	mad_fixed_t EqScale[32];

} libmad;

const datadef LibMadParams[] = 
{
	{ LIBMAD_ID,		TYPE_CLASS },
	{ LIBMAD_INPUT,		TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ LIBMAD_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },
	{ LIBMAD_COMMENTS,	TYPE_PIN, DF_OUTPUT, PACKET_COMMENT },
	{ LIBMAD_EQ,		TYPE_EQUALIZER, DF_HIDDEN },

	DATADEF_END
};

const char* Id3[] =
{
	ID3_FRAME_TITLE, "TITLE=",
	ID3_FRAME_ARTIST, "ARTIST=",
	ID3_FRAME_ALBUM,  "ALBUM=",
	NULL,NULL
};

static void UpdateScale( libmad* p )
{
	int i;
	for (i=0;i<32;++i)
		p->EqScale[i] = mad_f_mul(p->EqTable[i],p->Scale);
}

static int UpdateEq( libmad* p )
{
	static const uint8_t Map[32] = 
	{
		0, 1, 2, 3, 4, 5, 6, 6, 
		7, 7, 7, 7, 8, 8, 8, 8,
		8, 8, 8, 8, 9, 9, 9, 9, 
		9, 9, 9, 9, 9, 9, 9, 9
	};

	int i;
	for (i=0;i<32;++i)
	{
		int db = p->Eq.Amplify + p->Eq.Eq[Map[i]];
		if (db>18)
			db=18;
		p->EqTable[i] = mad_f_tofixed(pow(10,db/20.f));
	}

	p->Scale=MAD_F_ONE;
	UpdateScale(p);
	return ERR_NONE;
}

static int SetFormat( libmad* p, const audio* Format )
{
	if (p->InputFormat.Format)
	{
		memset(&p->InputFormat,0,sizeof(p->InputFormat));
		MergeDone(&p->Merge);
		mad_frame_finish(&p->Frame);
		mad_stream_finish(&p->Stream);
		mad_synth_finish(&p->Synth);
	}

	if (Format)
	{
		// accept only mp2 and mp3 
		if (Format->Format != AUDIOFMT_MP2 && Format->Format != AUDIOFMT_MP3)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		p->OutputFormat.Bits = sizeof(mad_fixed_t)*8;
		p->OutputFormat.Flags = PCM_PLANES;
		p->OutputFormat.Format = AUDIOFMT_PCM;
		p->OutputFormat.FracBits = MAD_F_FRACBITS;
		p->OutputFormat.SampleRate = p->InputFormat.SampleRate;	// optional
		p->OutputFormat.Channels = p->InputFormat.Channels; // optional

		p->Skip = 0;
		p->Pending = 0;
		p->FormatSet = 0;
		p->RefTime = 0;
		p->Total = 0;

		UpdateEq(p);

		MergeInit(&p->Merge);
		mad_stream_init(&p->Stream);
		mad_frame_init(&p->Frame);
		mad_synth_init(&p->Synth);
	}

	return ERR_NONE;
}

void TagParse( const void* Ptr, int Len, pin* Pin )
{
	struct id3_tag *Tag;
	Tag = id3_tag_parse(Ptr, Len);
	if (Tag) 
	{
		struct id3_frame const *Frame;
		union id3_field const *Field;
		id3_ucs4_t const *UCS4;
		id3_latin1_t *Latin1;
		int n,Count;
		const char** i;
		tchar_t Comment[256];

		for (i=Id3;i[0];i+=2)
		{
			Frame = id3_tag_findframe(Tag, i[0], 0);
			if (Frame)
			{
				Field = &Frame->fields[1];
				Count = id3_field_getnstrings(Field);

				for (n=0;n<Count;++n) 
				{
					UCS4 = id3_field_getstrings(Field,n);

					Latin1 = id3_ucs4_latin1duplicate(UCS4);
					if (Latin1)
					{
						strtotcs(Comment,i[1],256);
						strtotcs(Comment+tcslen(Comment),(char*)Latin1,256-tcslen(Comment));
						free(Latin1);

						Pin->Node->Set(Pin->Node,Pin->No,Comment,sizeof(Comment));
					}
				}
			}
		}

		id3_tag_delete(Tag);
		//mad_stream_skip(&p->Stream,TagSize);
	}
}

bool_t DecodeFrame( libmad* p )
{
	for (;;)
	{
		while (mad_frame_decode(&p->Frame, &p->Stream) == -1)
		{
			if (p->Stream.error == MAD_ERROR_LOSTSYNC && p->Comments.Node)
			{
				int TagSize = id3_tag_query(p->Stream.this_frame,p->Stream.bufend - p->Stream.this_frame);
				if (TagSize > 0) 
				{
					if (TagSize > p->Stream.bufend - p->Stream.this_frame)
					{
						MergeEnd(&p->Merge,p->Stream.this_frame - p->Stream.buffer);
						return 0;
					}

					TagParse(p->Stream.this_frame, TagSize, &p->Comments);
				}
			}
			else if (p->Stream.error == MAD_ERROR_BUFLEN || p->Stream.error == MAD_ERROR_BUFPTR)
			{
				MergeEnd(&p->Merge,p->Stream.next_frame - p->Stream.buffer);
				return 0;
			}
		}

		if (p->Eq.Enabled)
		{
			int n = MAD_NCHANNELS(&p->Frame.header);
			int m = MAD_NSBSAMPLES(&p->Frame.header);
			int ch;

			for (ch=0;ch<n;++ch)
			{
				mad_fixed_t* i;
				mad_fixed_t* ie;
				int eq=0;

				ie=p->Frame.sbsample[ch][m];
				for (i=p->Frame.sbsample[ch][0];i!=ie;i+=4,eq+=4)
				{
					eq &= 31;
					i[0] = mad_f_mul(i[0],p->EqScale[eq+0]);
					i[1] = mad_f_mul(i[1],p->EqScale[eq+1]);
					i[2] = mad_f_mul(i[2],p->EqScale[eq+2]);
					i[3] = mad_f_mul(i[3],p->EqScale[eq+3]);
				}
			}
		}

		mad_synth_frame(&p->Synth,&p->Frame);

		if (p->Skip)
		{
			// the first frame may be corrupted
			p->Skip = 0;
			continue;
		}

		// handle stereo streams with random mono frames (is this really a good mp3?)

		if (p->OutputFormat.Channels==2 && p->Synth.pcm.channels==1)
		{
			p->Synth.pcm.channels = 2;
			memcpy(p->Synth.pcm.samples[1],p->Synth.pcm.samples[0],
				p->Synth.pcm.length * sizeof(mad_fixed_t));
		}

		p->InputFormat.ByteRate = p->Frame.header.bitrate >> 3;

		// output format setup needed?

		if (p->OutputFormat.SampleRate != (int)p->Synth.pcm.samplerate ||
			p->OutputFormat.Channels != p->Synth.pcm.channels)
		{
			if (p->FormatSet)
			{
				p->FormatSet--;
				continue; // probably a bad frame, drop it
			}

			// set new output format
			p->InputFormat.SampleRate = p->OutputFormat.SampleRate = p->Synth.pcm.samplerate;
			p->InputFormat.Channels = p->OutputFormat.Channels = p->Synth.pcm.channels;
			
			if (p->Output.Node)
				p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));
		}
		p->FormatSet = 16; // reset countdown

		if (p->Eq.Enabled)
		{
			int Max = 0;
			int ch;

			for (ch=0;ch<p->Synth.pcm.channels;++ch)
			{
				mad_fixed_t* i;
				mad_fixed_t* ie;

				ie=p->Synth.pcm.samples[ch] + p->Synth.pcm.length;
				for (i=p->Synth.pcm.samples[ch];i!=ie;++i)
				{
					mad_fixed_t v = *i;
					if (v<0) v=-v;
					if (v>Max) Max=v;
				}
			}

			if (Max > MAD_F_ONE)
			{
				p->Scale = mad_f_tofixed(
					mad_f_todouble(p->Scale) /
					mad_f_todouble(MAD_F_ONE + mad_f_mul(Max - MAD_F_ONE, MAD_F(0x08000000))));

				UpdateScale(p);
			}
		}

		// build output packet
		p->Packet.RefTime = p->RefTime;
		p->Packet.Length = p->Synth.pcm.length * sizeof(mad_fixed_t);
		p->Packet.Data[0] = p->Synth.pcm.samples[0];
		p->Packet.Data[1] = p->Synth.pcm.samples[1];
		p->Total += p->Synth.pcm.length;

		// update output info
		p->RefTime = -1;
		return 1;
	}
}

static int Decode( libmad* p, const packet* Packet )
{
	int OldLength;
	int Result;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	if (p->Pending)
	{
		do
		{
			if (p->Output.Node->Set(p->Output.Node,p->Output.No,
									&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
				return ERR_BUFFER_FULL;
		}
		while (DecodeFrame(p));

		p->Pending = 0;
	}

	// add new packet to mergebuffer

	OldLength = p->Merge.Pos;
	MergeAdd(&p->Merge,Packet->Data[0],Packet->Length);

	// set new reftime (rewind with bytes already in mergebuffer)
	if (Packet->RefTime >= 0)
		p->RefTime = Packet->RefTime - Scale(OldLength,TICKSPERSEC,p->Frame.header.bitrate >> 3);

	mad_stream_buffer(&p->Stream, p->Merge.Data, p->Merge.Pos);

	if (DecodeFrame(p))
	{
		Result = ERR_NONE;
		do
		{
			int i = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));
			if (i == ERR_BUFFER_FULL)
			{
				p->Pending = 1;
				//return last result
				break;
			}

			Result = i;
		}
		while (DecodeFrame(p));
	}
	else
		Result = ERR_NEED_MORE_DATA; // if no frame decoded: need more packet

	return Result;
}

static bool_t IsEmpty( libmad* p )
{
	packet Zero;

	Zero.CurrTime = -1;
	Zero.DropLevel = p->Packet.DropLevel;
	Zero.Length = 0;
	Zero.Data[0] = NULL;
	Zero.RefTime = -1;

	return Decode(p,&Zero) == ERR_NEED_MORE_DATA;
}

static int Enum( libmad* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,LibMadParams,NULL);
}

static int Get( libmad* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(LIBMAD_ID,int); break;
	case LIBMAD_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case LIBMAD_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case LIBMAD_OUTPUT: GETVALUE(p->Output,pin); break;
	case LIBMAD_COMMENTS: GETVALUE(p->Comments,pin); break;
	case LIBMAD_EQ: GETVALUE(p->Eq,equalizer); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( libmad* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case LIBMAD_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case LIBMAD_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case LIBMAD_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case LIBMAD_COMMENTS: SETVALUE(p->Comments,pin,ERR_NONE); break;
	case LIBMAD_EQ: SETVALUE(p->Eq,equalizer,UpdateEq(p)); break;
	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_EMPTY:

		// reset
		MergeClear(&p->Merge);
		p->Pending = 0;
		p->Skip = 1;

		mad_frame_finish(&p->Frame);
		mad_synth_finish(&p->Synth);

		mad_frame_init(&p->Frame);
		mad_synth_init(&p->Synth);
		break;
	}
	return Result;
}

static libmad LibMad = 
{
	Enum,
	Get,
	Set,
};

void LibMad_Init()
{
	NodeRegister(&LibMad.VMT,PRI_DEFAULT);
}

void LibMad_Done()
{
	NodeUnRegister(&LibMad.VMT);
}



