/*****************************************************************************
 *
 * This program is free software ; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * $Id$
 *
 ****************************************************************************/

#include "../common/common.h"
#include "vorbis.h"
#include <math.h>
#include "tremor/ogg.h"
#include "tremor/ivorbiscodec.h"

// Vorbis audio codec

#define VORBIS_INPUT		0x100
#define VORBIS_OUTPUT		0x101
#define VORBIS_EQ			0x102

typedef struct vorbis
{
	node VMT;

	audio InputFormat;
	audio OutputFormat;
	pin Output;

	equalizer Eq;

	vorbis_dsp_state DSP;
	vorbis_block     Block;

	// output data
	packet Packet;
	int Samples; //samples pending
	int Total;

	ogg_int32_t Scale;
	ogg_int32_t EqAmp;
	ogg_int32_t EqScale;

} vorbis;

const datadef VorbisParams[] = 
{
	{ VORBIS_ID,		TYPE_CLASS },
	{ VORBIS_INPUT,		TYPE_PACKET, DF_INPUT, PACKET_AUDIO },
	{ VORBIS_OUTPUT,	TYPE_PIN, DF_OUTPUT, PACKET_AUDIO },
	{ VORBIS_EQ,		TYPE_EQUALIZER, DF_HIDDEN },

	DATADEF_END
};

static void UpdateScale( vorbis* p )
{
	p->EqScale = Scale(p->Scale,p->EqAmp,1<<24);
}

static int UpdateEq( vorbis* p )
{
	int db = p->Eq.Amplify;
	if (db>18)
		db=18;
	p->EqAmp = (int)(pow(10,db/20.f)*(1<<24));
	
	p->Scale = 1<<24;
	UpdateScale(p);
	return ERR_NONE;
}

static int SetFormat( vorbis* p, const audio* Format )
{
	if (p->InputFormat.Format)
	{
		memset(&p->InputFormat,0,sizeof(p->InputFormat));

		vorbis_block_clear(&p->Block);
		vorbis_dsp_clear(&p->DSP);
	}

	if (Format)
	{
		vorbis_info* Info = (vorbis_info*) Format->User;

		// accepts only voribs (ogg_packet)
		if (Format->Format != AUDIOFMT_VORBIS_INTERNAL)
			return ERR_INVALID_DATA;

		p->InputFormat = *Format;

		p->OutputFormat.Bits = sizeof(ogg_int32_t)*8;
		p->OutputFormat.Flags = PCM_PLANES;
		p->OutputFormat.Format = AUDIOFMT_PCM;
		p->OutputFormat.FracBits = 24;

		p->Samples = 0;
		p->Total = 0;
		UpdateEq(p);

		if (Info)
		{
			vorbis_synthesis_init(&p->DSP,Info);
			vorbis_block_init(&p->DSP,&p->Block);

			p->InputFormat.ByteRate = Info->bitrate_nominal >> 3;
			p->InputFormat.SampleRate = p->OutputFormat.SampleRate = Info->rate;
			p->InputFormat.Channels = p->OutputFormat.Channels = Info->channels;

			if (p->Output.Node)
				p->Output.Node->Set(p->Output.Node,p->Output.No|PACKET_FORMAT,&p->OutputFormat,sizeof(audio));
		}
	}

	return ERR_NONE;
}

static int Decode( vorbis* p, const packet* Packet )
{
	int Result = ERR_NONE;

	p->Packet.CurrTime = Packet->CurrTime;
	p->Packet.DropLevel = Packet->DropLevel;

	// pending packet?

	if (p->Samples)
	{
		if (p->Output.Node->Set(p->Output.Node,p->Output.No,
								&p->Packet,sizeof(packet)) == ERR_BUFFER_FULL)
			return ERR_BUFFER_FULL;

		vorbis_synthesis_read(&p->DSP,p->Samples);
		p->Samples = 0;
	}

	if (Packet->Length != sizeof(ogg_packet))
		return ERR_INVALID_DATA;

	if (vorbis_synthesis(&p->Block,(ogg_packet*)Packet->Data[0],1)==0)
	{
		ogg_int32_t **PCM;

		vorbis_synthesis_blockin(&p->DSP,&p->Block);

		p->Samples = vorbis_synthesis_pcmout(&p->DSP,&PCM);

		if (p->Samples)
		{
			if (p->Eq.Enabled)
			{
				int Max = 0;
				int ch;

				for (ch=0;ch<p->OutputFormat.Channels;++ch)
				{
					ogg_int32_t* i;
					ogg_int32_t* ie;
					ogg_int32_t s = p->EqScale >> 16;

					ie=PCM[ch] + p->Samples;
					for (i=PCM[ch];i!=ie;++i)
					{
						ogg_int32_t v = (*i >> 8)*s;
						*i = v;
						if (v<0) v=-v;
						if (v>Max) Max=v;
					}
				}

				if (Max > (1<<24))
				{
					p->Scale = Scale(p->Scale,1<<24,(1<<23) + (Max >> 1));
					UpdateScale(p);
				}
			}

			// build output packet
			p->Packet.RefTime = Packet->RefTime;
			p->Packet.Length = p->Samples * sizeof(ogg_int32_t);
			p->Packet.Data[0] = PCM[0];
			p->Packet.Data[1] = PCM[1];
			p->Total += p->Samples;

			Result = p->Output.Node->Set(p->Output.Node,p->Output.No,
						 				 &p->Packet,sizeof(packet));

			if (Result != ERR_BUFFER_FULL)
			{
				vorbis_synthesis_read(&p->DSP,p->Samples);
				p->Samples = 0;
			}
			else
				Result = ERR_NONE; // don't return buffer full. packet already processed
		}
	}
	
	return Result;
}

static bool_t IsEmpty( vorbis* p )
{
	if (p->Samples)
	{
		packet Zero;

		Zero.CurrTime = -1;
		Zero.DropLevel = p->Packet.DropLevel;
		Zero.Length = 0;
		Zero.RefTime = -1;

		Decode(p,&Zero);
	}

	return !p->Samples;
}

static int Enum( vorbis* p, int No, datadef* Param )
{
	return NodeEnumType(&No,Param,NodeParams,FlowParams,VorbisParams,NULL);
}

static int Get( vorbis* p, int No, void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case NODE_ID: GETVALUE(VORBIS_ID,int); break;
	case VORBIS_INPUT | PACKET_FORMAT: GETVALUE(p->InputFormat,audio); break;
	case VORBIS_OUTPUT | PACKET_FORMAT: GETVALUE(p->OutputFormat,audio); break;
	case VORBIS_OUTPUT: GETVALUE(p->Output,pin); break;
	case VORBIS_EQ: GETVALUE(p->Eq,equalizer); break;
	case FLOW_TOTAL:GETVALUE(p->Total,int); break;
	case FLOW_EMPTY:GETVALUE(IsEmpty(p),bool_t); break;
	}
	return Result;
}

static int Set( vorbis* p, int No, const void* Data, int Size )
{
	int Result = ERR_INVALID_PARAM;
	switch (No)
	{
	case VORBIS_INPUT | PACKET_FORMAT:
		if (!Size || Size == sizeof(audio))
		{
			if (!Size) Data = NULL;
			Result = SetFormat(p,(const audio*)Data);
		}
		break;

	case VORBIS_INPUT:
		if (Size == sizeof(packet))
			Result = Decode(p,(const packet*)Data);
		break;

	case VORBIS_OUTPUT:
		if (Size == sizeof(pin))
			Result = NodeSetPin(&p->Output,(pin*)Data,&p->OutputFormat,sizeof(audio));
		break;

	case VORBIS_EQ: SETVALUE(p->Eq,equalizer,UpdateEq(p)); break;
	case FLOW_TOTAL: SETVALUE(p->Total,int,ERR_NONE); break;
	case FLOW_EMPTY:
		vorbis_synthesis_restart(&p->DSP);
		p->Samples = 0;
		break;
	}
	return Result;
}

static vorbis Vorbis = 
{
	Enum,
	Get,
	Set,
};

void Vorbis_Init()
{
	NodeRegister((node*)&Vorbis.VMT,PRI_DEFAULT);
}

void Vorbis_Done()
{
	NodeUnRegister((node*)&Vorbis.VMT);
}
